"""A class for crossmatching clusters with literature catalogues."""

import numpy as np
import pandas as pd

from ..calculate.constants import default_ocelot_key_names
from astropy import units as u
from astropy.coordinates import SkyCoord
from scipy.stats import norm
from typing import Optional, Union, List, Tuple


class Catalogue:
    # Todo this class needs refactoring and overhauling. It's... messy...
    def __init__(
        self,
        data: pd.DataFrame,
        catalogue_name: str,
        key_name: str = "Name",
        key_ra: str = "ra",
        key_dec: str = "dec",
        key_tidal_radius: str = "ang_radius_t",
        extra_axes: Optional[Union[list, tuple, np.ndarray]] = None,
        assumed_position_systematic_error: float = 0.0,
        match_tidal_radius: bool = False,
    ):
        """Storage system for existing catalogues that handles cross-matching too.

        Args:
            data (pd.DataFrame or pd.Series if just one cluster): data for the catalogue to match against.
            catalogue_name (str): name of the catalogue. This is stored and used to
            key_name (str): name of the name column in data.
                Default: "Name"
            key_ra (str): name of the ra column in data.
                Default: "ra"
            key_dec (str): name of the dec column in data.
                Default: "dec"
            extra_axes (list-like, optional): array of extra axes to match against in shape (n_axes, 2), where the
                array elements give the following names:
                0: name of the column in data
                1: name of the error column in data - may simply be "None"
            assumed_position_systematic_error (float): assumed Gaussian systematic error on the position of stars in the
                catalogue. Useful when catalogues have positional errors to weight the closeness between matched and
                catalogue clusters.
                Default: 0.0, although this is *rarely zero* as many catalogues will do a shitty job of inferring the
                center of clusters!
            match_tidal_radius (bool): whether or not to also match with tidal radii. We'll specify the angular
                separation in terms of tidal radius between literature & found, too.
                Default: False

        """
        # Initialise the catalogue's name, cluster names and co-ordinates
        self.catalogue_name = catalogue_name
        self.names = data[key_name].to_numpy()
        self.coords = SkyCoord(
            ra=data[key_ra].to_numpy() << u.deg, dec=data[key_dec].to_numpy() << u.deg
        )
        self.assumed_position_systematic_error = assumed_position_systematic_error

        self.match_tidal_radius = match_tidal_radius
        if match_tidal_radius:
            self.tidal_radius_data = data[key_tidal_radius].to_numpy()

        # Read in any extra axes
        if extra_axes is None:
            self.n_extra_features = 0
            self.extra_axes_catalogue_column_names = None
            self.extra_axes_data = None
            self.extra_axes_data_errors = None
        else:
            self.n_extra_features = len(extra_axes)
            self.extra_axes_catalogue_column_names = [None] * self.n_extra_features
            self.extra_axes_data = [None] * self.n_extra_features
            self.extra_axes_data_errors = [None] * self.n_extra_features

            # Cycle over the axes, adding the data to the lists in np.ndarray format
            for i, an_axis in enumerate(extra_axes):
                self.extra_axes_catalogue_column_names[i] = an_axis[0]
                self.extra_axes_data[i] = np.asarray(data[an_axis[0]])

                # Deal with a possible lack of specified error information
                if an_axis[1] is None:
                    self.extra_axes_data_errors[i] = np.zeros(data.shape[0])
                else:
                    self.extra_axes_data_errors[i] = np.asarray(data[an_axis[1]])

        # Write the usual package-wide keys for clusters to the class
        self.ocelot_key_names = default_ocelot_key_names

    def override_default_ocelot_parameter_names(
        self, parameters_to_override: dict
    ) -> None:
        """A function for overriding the default parameter names assigned to open clusters by ocelot.calculate.
        Necessary for crossmatching with a catalogue produced by another module.

        Args:
            parameters_to_override (dict of str): parameters to override, in the format "parameter": "new_name". See
                ocelot.calculate.constants.default_ocelot_key_names for a list of all parameters that can be
                overwriten.

        Returns:
            None

        """
        for a_parameter in parameters_to_override:
            self.ocelot_key_names[a_parameter] = parameters_to_override[a_parameter]

    def _two_parameter_position_crossmatch(
        self, clusters: SkyCoord, maximum_separation: float
    ):
        """Todo fstring"""

        # Perform the matching
        (
            id_clusters,
            id_catalog,
            distances_2d,
            distances_3d,
        ) = self.coords.search_around_sky(clusters, maximum_separation * u.deg)

        return id_clusters, id_catalog, distances_2d

    @staticmethod
    def _calculate_sigma(
        crossmatched_value: float,
        catalog_value: float,
        random_error: float,
        systematic_error: float = 0.0,
    ):
        """A function for calculating sigma difference between a literature object and a crossmatching object. Can
        also consider uniformly distributed systematic errors."""
        # We calculate the difference, then reduce that difference by the systematic error, make sure the sigma is
        # still >= 0.0 (sigma=0.0 means difference is systematic-dominated) and then divide by the random error toooo
        return (
            np.clip(
                np.abs(catalog_value - crossmatched_value) - systematic_error,
                0.0,
                np.inf,
            )
            / random_error
        )

    def crossmatch(
        self,
        data_cluster: pd.DataFrame,
        extra_axes_keys: Union[Tuple[str], List[str]] = (),
        max_separation: float = 1.0,
        best_match_on: str = "mean_sigma",
        tidal_radius_mode: str = "max",
        matches_to_record: int = 2,
        max_sigma_threshold: float = 5,
        position_systematics: float = 0.0,
        extra_axes_systematics: Optional[Union[list, tuple, np.ndarray]] = None,
    ):
        """Crossmatches the stored catalogue with one/multiple clusters depending on what was specified by the
        user.

        Notes:
            - Typically, matches should be ordered best based on their mean sigma (e.g. a cluster with sigmas 0, 2 is a
                better match than one with 1.5, 1.5 sigmas), but should be decided matched or not based on their max
                sigma (a cluster with a large max sigma simply isn't a match.)

        Args:
            data_cluster (pd.DataFrame): data on the clusters to match against, in the shape (n_clusters, n_axes).
            extra_axes_keys (list, optional): a list of the names of extra axes to match against. Should be keys that
                can be found in ocelot.calculate.constants.default_ocelot_key_names. Default names can be overriden
                or added to with override_default_ocelot_parameter_names.
            max_separation (float in degrees): maximum separation on the sky to report matches for.
                Default: 1 degree (which should be ok most of the time)
            best_match_on (str): what to make the best match on. May use "just_position" or may use "max_sigma" or
                "mean_sigma" of all axes.
                Default: "mean_sigma"
            tidal_radius_mode (str): which tidal radius value to use or how. Allowed values:
                - "literature": uses the literature value only
                - "data":       uses the data value only
                - "max":        uses the max value of the two
                - "mean":       uses the mean value of the two
                Default: "max"
            matches_to_record (int): the maximum number of matches to report back for a given cluster.
                Default: 2
            max_sigma_threshold (float): the threshold maximum sigma level above which matches will not be considered
                a valid match.
                Default: 5.

        Returns:
            matches: a pd.DataFrame of all found matches, including data on the separations and probabilities of
                matches being true, of shape (n_matches, ? - the number of stats per match depends on extra_axes)
            match_statistics: a pd.DataFrame of all input clusters' total numbers of matches, useful for quick
                diagnostics of a crossmatch run, of shape (n_input_clusters, ?)

        """
        # -------------------------------------
        # INITIAL SETUP AND READING IN OF DATA
        # Check that the correct number of extra_axes_keys have been specified based on the initialisation
        if len(extra_axes_keys) != self.n_extra_features:
            raise ValueError(
                "length of extra_axes_keys does not match the number of extra features specified at "
                "initialisation of the catalogue."
            )

        # Setup the extra axes systematics frames
        if extra_axes_systematics is None:
            extra_axes_systematics = np.zeros(self.n_extra_features)

        # Grab a list of extra axes key names from the class' ocelot parameter dict
        # We'll store main parameters in even indices (0, 2, 4) and their errors at the next odd one (1, 3, 5)...
        extra_axes_ocelot_key_names = []
        for a_parameter in extra_axes_keys:
            extra_axes_ocelot_key_names.append(self.ocelot_key_names[a_parameter])
            extra_axes_ocelot_key_names.append(
                self.ocelot_key_names[a_parameter + "_error"]
            )

        # Grab all the data we need from data_cluster
        names = np.asarray(data_cluster[self.ocelot_key_names["name"]])
        ra = np.asarray(data_cluster[self.ocelot_key_names["ra"]])
        dec = np.asarray(data_cluster[self.ocelot_key_names["dec"]])
        extra_axes = np.asarray(data_cluster[extra_axes_ocelot_key_names])

        # -------------------------------------
        # 2D ON-SKY CROSSMATCHING
        # Gather all possible 2D matches below the threshold max_separation:
        cluster_skycoords = SkyCoord(ra=ra << u.deg, dec=dec << u.deg)
        id_clusters, id_catalog, distances_2d = self._two_parameter_position_crossmatch(
            cluster_skycoords, max_separation
        )

        # -------------------------------------
        # CREATION OF MATCH DATA DATAFRAME FOR POSITION
        # Turn this into a fancy DataFrame thing
        match_data = pd.DataFrame(
            {
                "name": names[id_clusters],
                "name_match": self.names[id_catalog],
                "angular_sep": distances_2d,
            }
        )

        # Calculate the error in the position estimate - firstly, find the positional errors on potential clusters in
        # ra, dec
        ra_error = np.sqrt(
            np.asarray(data_cluster.loc[id_clusters, self.ocelot_key_names["ra_error"]])
            ** 2
            + self.assumed_position_systematic_error**2
        )
        dec_error = np.sqrt(
            np.asarray(
                data_cluster.loc[id_clusters, self.ocelot_key_names["dec_error"]]
            )
            ** 2
            + self.assumed_position_systematic_error**2
        )

        # We only find the sigma values where the error isn't zero - else we get division by 0 =(
        # This should never happen, but we do the check here anyway!
        good_ra = ra_error != 0
        ra_sigma = np.empty(ra_error.shape)
        ra_sigma[good_ra] = self._calculate_sigma(
            ra[id_clusters],
            self.coords.ra.value[id_catalog],
            ra_error,
            systematic_error=position_systematics,
        )
        ra_sigma[np.invert(good_ra)] = 0.0

        good_dec = dec_error != 0
        dec_sigma = np.empty(dec_error.shape)
        dec_sigma[good_dec] = self._calculate_sigma(
            dec[id_clusters],
            self.coords.dec.value[id_catalog],
            dec_error,
            systematic_error=position_systematics,
        )
        dec_sigma[np.invert(good_dec)] = 0.0

        # Lastly, we can grab the total sigma values and convert this into a probability
        match_data["angular_sep_sigma"] = np.sqrt(ra_sigma**2 + dec_sigma**2)
        match_data["angular_sep_prob"] = 2 * norm.cdf(
            -match_data["angular_sep_sigma"], loc=0.0, scale=1.0
        )

        # And also do some stuff with the tidal radii
        if self.match_tidal_radius:
            radius_literature = self.tidal_radius_data[id_catalog]
            radius_data = data_cluster[
                self.ocelot_key_names["ang_radius_t"]
            ].to_numpy()[id_clusters]

            # Decide on which form to use
            if tidal_radius_mode == "literature":
                tidal_radii = radius_literature
            elif tidal_radius_mode == "data":
                tidal_radii = radius_data
            elif tidal_radius_mode == "mean":
                tidal_radii = np.mean(
                    np.vstack([radius_literature, radius_data]), axis=0
                )
            elif tidal_radius_mode == "max":
                tidal_radii = np.maximum(radius_literature, radius_data)
            elif tidal_radius_mode == "min":
                tidal_radii = np.minimum(radius_literature, radius_data)
            else:
                raise ValueError(
                    f"specified tidal_radius_mode '{tidal_radius_mode}' not recognised/supported!"
                )

            match_data["tidal_sep_ratio"] = match_data["angular_sep"] / tidal_radii
            match_data["tidal_sep_ratio_literature"] = (
                match_data["angular_sep"] / radius_literature
            )
            match_data["tidal_sep_ratio_data"] = match_data["angular_sep"] / radius_data

        # -------------------------------------
        # EXTENSION OF MATCH DATA FRAME TO THE EXTRA AXES
        # Incrementally work through the extra axes
        list_of_sigmas = []
        i_axis = 0
        while i_axis < self.n_extra_features:
            # Get the combined error with quadrature
            combined_error = np.sqrt(
                (self.extra_axes_data_errors[i_axis])[id_catalog] ** 2
                + extra_axes[id_clusters, 2 * i_axis + 1] ** 2
            )

            # Make axis names
            an_axis_name = extra_axes_keys[i_axis] + "_sep"
            an_axis_name_sigma = an_axis_name + "_sigma"
            list_of_sigmas.append(an_axis_name_sigma)

            # Calculate and store the separation
            match_data[an_axis_name] = np.abs(
                (self.extra_axes_data[i_axis])[id_catalog]
                - extra_axes[id_clusters, 2 * i_axis]
            )

            # Quantify the separation in terms of the error
            match_data[an_axis_name_sigma] = self._calculate_sigma(
                extra_axes[id_clusters, 2 * i_axis],
                (self.extra_axes_data[i_axis])[id_catalog],
                combined_error,
                systematic_error=extra_axes_systematics[i_axis],
            )

            i_axis += 1

        # Make some summary statistics that use all of the above
        if self.n_extra_features != 0 and best_match_on != "just_position":
            # angular_sep_sigma is irrelevant when working instead with tidal constraints, so we aren't interesting in
            # them.
            if best_match_on == "just_tidal_separation":
                match_data["max_sigma"] = np.max(match_data[list_of_sigmas], axis=1)
                match_data["mean_sigma"] = np.mean(match_data[list_of_sigmas], axis=1)

            else:
                match_data["max_sigma"] = np.max(
                    match_data[list_of_sigmas + ["angular_sep_sigma"]], axis=1
                )
                match_data["mean_sigma"] = np.mean(
                    match_data[list_of_sigmas + ["angular_sep_sigma"]], axis=1
                )
        else:
            match_data["max_sigma"] = match_data["angular_sep_sigma"]
            match_data["mean_sigma"] = match_data["angular_sep_sigma"]

        # -------------------------------------
        # CREATION OF SUMMARY DATAFRAME
        # See which clusters have matches in the matches returned
        cluster_has_a_match = pd.Series(names).isin(match_data["name"])
        names_of_clusters_with_matches = pd.Series(names)[cluster_has_a_match]

        # Also make an array of the total number of matches per cluster
        matches_per_cluster = np.zeros(names.shape, dtype=int)
        matches_per_cluster[cluster_has_a_match] = (match_data["name"].value_counts())[
            names_of_clusters_with_matches
        ]

        # Set which column we're gonna use to pick best matches
        # NOTE TO FUTURE ME: this should always be done with sigma values - *not* probabilities - as the best match is
        # then found later by sorting in *ascending* order. I.e. the "best thing" in a column must be the minimum thing.
        if best_match_on == "max_sigma":
            best_match_column = "max_sigma"
        elif best_match_on == "mean_sigma":
            best_match_column = "mean_sigma"
        elif best_match_on == "just_position":
            best_match_column = "angular_sep_sigma"
        elif best_match_on == "just_tidal_separation":
            best_match_column = "tidal_sep_ratio"
        else:
            raise ValueError(
                "specified best_match_on invalid: may only be one of 'max_sigma', 'mean_sigma', "
                "'just_position' or 'just_tidal_separation'."
            )

        # Create a summary DataFrame with the best matches for each cluster
        summary_match_data = pd.DataFrame(
            {
                "name": names,
                "valid_matches": np.zeros(matches_per_cluster.shape, dtype=int),
                "total_matches": matches_per_cluster,
            }
        )
        i_match = 0
        while i_match < matches_to_record:
            summary_match_data[f"match_{i_match}"] = np.nan
            summary_match_data[f"match_{i_match}_angular_sep"] = np.nan
            if self.match_tidal_radius:
                summary_match_data[f"match_{i_match}_tidal_sep_ratio"] = np.nan
            summary_match_data[f"match_{i_match}_max_sigma"] = np.nan
            summary_match_data[f"match_{i_match}_mean_sigma"] = np.nan
            i_match += 1

        # Cycle over clusters with matches, storing things about their best matches
        if self.match_tidal_radius:
            columns_to_read = [
                "name_match",
                "angular_sep",
                "tidal_sep_ratio",
                "max_sigma",
                "mean_sigma",
            ]
        else:
            columns_to_read = ["name_match", "angular_sep", "max_sigma", "mean_sigma"]

        for i_cluster in names_of_clusters_with_matches.index:
            # Make a new DataFrame of the current matches and get a sorted list of the best IDs
            current_matches = match_data.loc[
                match_data["name"] == names_of_clusters_with_matches[i_cluster], :
            ].reset_index(drop=True)
            best_match_ids = (
                current_matches[best_match_column].sort_values(ascending=True).index
            )

            # Count the number of matches below the sigma threshold
            summary_match_data.loc[i_cluster, "valid_matches"] = np.count_nonzero(
                np.asarray(current_matches["max_sigma"] < max_sigma_threshold)
            )

            # Add these matches to the main DataFrame
            i_match = 0
            while i_match < np.min([matches_to_record, len(best_match_ids)]):
                # Boilerplate to get the ID of the match and the prefix of the columns to write to
                an_id = best_match_ids[i_match]
                match = f"match_{i_match}"

                # And now, write the data to the overall DataFrame
                if self.match_tidal_radius:
                    columns_to_write = [
                        match,
                        match + "_angular_sep",
                        match + "_tidal_sep_ratio",
                        match + "_max_sigma",
                        match + "_mean_sigma",
                    ]
                else:
                    columns_to_write = [
                        match,
                        match + "_angular_sep",
                        match + "_max_sigma",
                        match + "_mean_sigma",
                    ]

                summary_match_data.loc[
                    i_cluster, columns_to_write
                ] = current_matches.loc[an_id, columns_to_read].values
                i_match += 1

        return match_data, summary_match_data



"""Set of functions for reading in isochrones from their downloaded formats and processing them into useful spaces."""

from pathlib import Path
from typing import Union, Optional, List

import numpy as np
import pandas as pd
from astropy.io import ascii


def read_parsec(
    file_location: Union[Path, List[Path], str, List[str]],
    max_label: int = 9,
    column_names: Optional[str] = None,
) -> pd.DataFrame:
    """Reads input CMD 3.3 isochrone(s). In the plural case, they're combined into one 
    DataFrame.

    Only verified for use when reading in isochrones from the CMD v3.3 & PARSEC v1.2S 
    web interface at http://stev.oapd.inaf.it/cgi-bin/cmd.

    Parameters
    ----------
    file_location : pathlib.Path or list of pathlib.Path or str
        You can specify three things:
        - Location of a file containing CMD 3.3 / PARSEC v1.2s isochrones.
        - A list of files as the above - they will all be joined.
        - A directory of .dat files containing isochrones
    max_label : int
        Maximum label to read in, which defaults to 9. The upper end of PARSEC tracks 
        can be a little bit experimental, so you may not want all labels (7 is a good 
        choice.) Default is 9 (the highest). See notes for more details.  
    column_names : ArrayLike, optional
        Alternative column names to use. Default: None, which uses column names:
        ['Zini', 'MH', 'logAge', 'Mini', 'int_IMF', 'Mass', 'logL', 'logTe', 'logg', 
        'label', 'mbolmag', 'Gmag', 'G_BPmag', 'G_RPmag'] as from the CMD 3.3 output in 
        the Gaia DR2 Evans+2018 photometric system.

    Returns
    -------
    isochrone : pd.Dataframe
        a pd.DataFrame of the read-in isochrone. It's worth checking manually that this 
        worked, as the tables are in a format that requires some cleaning (that should 
        hopefully be done automatically.) It will also have had the colour calculated, 
        which will be labelled 'G_BP-RP'.

    Notes
    -----
    PARSEC v1.2s labels correspond to...
    - 0 = PMS, pre main sequence
    - 1 = MS, main sequence
    - 2 = SGB, subgiant branch, or Hertzsprung gap for more intermediate+massive stars
    - 3 = RGB, red giant branch, or the quick stage of red giant for 
        intermediate+massive stars
    - 4 = CHEB, core He-burning for low mass stars, or the initial stage of CHeB for 
        intermediate+massive stars
    - 5 = still CHEB, the blueward part of the Cepheid loop of intermediate+massive 
        stars
    - 6 = still CHEB, the redward part of the Cepheid loop of intermediate+massive stars
    - 7 = EAGB, the early asymptotic giant branch, or a quick stage of red giant for 
        massive stars
    - 8 = TPAGB, the thermally pulsing asymptotic giant branch
    - 9 = post-AGB (in preparation!)
    """
    # Todo refactor to be more cross-compatible with other PARSEC versions (and be generally better)
    # Use default column names if none are specified
    if column_names is None:
        column_names = [
            "Zini",
            "MH",
            "logAge",
            "Mini",
            "int_IMF",
            "Mass",
            "logL",
            "logTe",
            "logg",
            "label",
            "mbolmag",
            "Gmag",
            "G_BPmag",
            "G_RPmag",
        ]

    if isinstance(file_location, str):
        file_location = Path(file_location)

    # See if file_location is a directory - only works if one path is specified
    try:
        file_location_is_a_directory = file_location.is_dir()
    except AttributeError:  # Catches if file_location is a list of paths instead, and does not have an is_dir() method
        file_location_is_a_directory = False

    # Iterate over all files if a list of files or a directory has been specified
    if isinstance(file_location, list) or file_location_is_a_directory:
        # If a directory was specified, we want to first make a list of all its .dat files.
        if file_location_is_a_directory:
            file_location = list(file_location.glob("*.dat"))

        # Cycle over all the isochrones, combining them into a file
        list_of_isochrones = []
        for a_file in file_location:
            current_isochrones = ascii.read(str(a_file.resolve())).to_pandas()
            current_isochrones.columns = column_names
            list_of_isochrones.append(current_isochrones)

        # Combine all the isochrones into one and reset the indexing
        isochrones = pd.concat(list_of_isochrones, ignore_index=True)

    # Otherwise, just read in the one file
    else:
        isochrones = ascii.read(str(file_location.resolve())).to_pandas()
        isochrones.columns = column_names

    # Use default column names if none are specified
    if column_names is None:
        isochrones.columns = [
            "Zini",
            "MH",
            "logAge",
            "Mini",
            "int_IMF",
            "Mass",
            "logL",
            "logTe",
            "logg",
            "label",
            "mbolmag",
            "Gmag",
            "G_BPmag",
            "G_RPmag",
        ]

    # Drop anything that isn't the right type of star
    rows_with_bad_label = np.where(isochrones["label"] > max_label)[0]
    isochrones = isochrones.drop(labels=rows_with_bad_label, axis="index").reset_index(
        drop=True
    )

    # Add some extra useful things
    isochrones["G_BP-RP"] = isochrones["G_BPmag"] - isochrones["G_RPmag"]

    return isochrones


def read_mist():
    # Todo
    pass


"""A set of tools for post-processing data ready for plotting."""

from typing import Optional, Union

import numpy as np
from sklearn.neighbors import KernelDensity


def kde_fit_2d(xy: np.ndarray,
               kde_bandwidth: float,
               x_limits: Optional[Union[list, np.ndarray]] = None,
               y_limits: Optional[Union[list, np.ndarray]] = None,
               kde_resolution: int = 50):
    """Fits a Kernel Density Estimator (KDE) in 2 dimensions with a Gaussian
    kernel of specified bandwidth.

    # Todo: This could easily be generalised to be multi-dimensional, simply by using the dimensions of xy as a hint.

    Notes:
        Do not use x and y limits to restrict the data set - only to reduce the
        area for plotting output. A fit will still be performed across the
        entire data range, which could get very unoptimised if you only need to
        fit a small segment of the dataset!

        KDE runtime scales as O(nm), where n is the number of samples and m is
        the number of points to evaluate at. Reducing the dataset size or the
        resolution of scoring will result in faster running.

    Args:
        xy (np.ndarray): the x and y data, of shape (n_samples, n_features).
        kde_bandwidth: the bandwidth to be used by the KDE. Recommended: about 0.08.
        x_limits (list-like[floats], optional): x limits of the final KDE scoring grid.
        y_limits (list-like[floats], optional): y limits of the final KDE scoring grid.
        kde_resolution (int): number of points to sample the KDE at when scoring
            (across a resolution x resolution-sized grid.) Default: 50.

    Returns:
        [2D array of x values of scores,
         2D array of y values of scores,
         2D array of scores]

    """
    # Use minimum and maximum values if limits haven't been specified.
    if x_limits is None:
        x_limits = np.array([xy[:, 0].min(), xy[:, 0].max()])

    if y_limits is None:
        y_limits = np.array([xy[:, 1].min(), xy[:, 1].max()])

    # Make up a mesh grid based on the limits of the dataset, so we can do some
    # density plotting in a moment
    grid_x = np.linspace(x_limits[0], x_limits[1], kde_resolution)
    grid_y = np.linspace(y_limits[0], y_limits[1], kde_resolution)
    mesh_x, mesh_y = np.meshgrid(grid_x, grid_y)

    # Fit the KDE
    xy_kde = KernelDensity(bandwidth=kde_bandwidth,
                           metric='euclidean',
                           kernel='gaussian',
                           algorithm='ball_tree')
    xy_kde.fit(xy)

    # Evaluate the KDE on the meshes, inputting an argument of shape
    # (n_samples, n_features) to the sklearn kde
    z = xy_kde.score_samples(np.array([mesh_x.flatten(), mesh_y.flatten()]).T)

    # Re-normalise as the scores returned are logarithms
    z = np.exp(z)

    # Re-shape the scores to the right shape again
    z = z.reshape(mesh_x.shape)

    return [mesh_x, mesh_y, z]



"""Generic calculation utilities used across the module."""
import numpy as np
from numpy.typing import ArrayLike, NDArray
from typing import Optional, Union
from ocelot.util.check import _check_matching_lengths_of_non_nones
from scipy.stats import directional_stats


def _weighted_standard_deviation(x: ArrayLike, weights: Optional[ArrayLike] = None):
    """Computes weighted standard deviation. Uses method from
    https://stackoverflow.com/a/52655244/12709989.
    """
    # Todo: not sure that this deals with small numbers of points correctly!
    #   See: unit test fails when v. few points used
    return np.sqrt(np.cov(x, aweights=weights))


def standard_error(
    standard_deviation: Union[ArrayLike[Union[float, int]], float, int],
    number_of_measurements: Union[ArrayLike[int], int],
) -> float:
    """Calculates the standard error on the mean of some parameter given the standard
    deviation.

    Parameters
    ----------
    standard_deviation : array-like, float, or int
        Standard deviation(s)
    number_of_measurements : array-like of ints, int
        Number of measurements used to find standard deviation.

    Returns
    -------
    standard_error : float
    """
    return standard_deviation / np.sqrt(number_of_measurements)


def mean_and_deviation(
    values: ArrayLike,
    weights: Optional[ArrayLike] = None,
) -> tuple[float]:
    """Calculates the mean and standard deviation of some set of values.

    Parameters
    ----------
    values : array-like
        Values to calculate mean and standard deviation of.
    weights : array-like, optional
        Array of weights to use to compute a weighted mean and average.

    Returns
    -------
    mean : float
        Mean of values.
    std : float
        Standard deviation of values.
    """
    _check_matching_lengths_of_non_nones(values, weights)

    return (
        np.average(values, weights=weights),
        _weighted_standard_deviation(values, weights),
    )


def lonlat_to_unitvec(longitudes: NDArray, latitudes: NDArray):
    """Converts longitudes and latitudes to unit vectors on a unit sphere. Uses method 
    at https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates.
    Assumes that latitudes is in the range [-pi / 2, pi / 2] as is common in 
    astronomical unit systems.
    """
    x = np.cos(latitudes) * np.cos(longitudes)
    y = np.cos(latitudes) * np.sin(longitudes)
    z = np.sin(latitudes)
    return np.column_stack((x, y, z))


def unitvec_to_lonlat(unit_vectors: NDArray):
    """Converts unit vectors on a unit sphere to longitudes and latitudes. See 
    `lonlat_to_unitvec` for more details.
    """
    x, y, z = [column.ravel() for column in np.hsplit(unit_vectors, 3)]
    longitudes = np.arctan2(y, x)
    latitudes  = np.arcsin(z / np.sqrt(x**2 + y**2 + z**2))
    return longitudes, latitudes


def spherical_mean(longitudes: ArrayLike, latitudes: ArrayLike):
    """Calculates the spherical mean of angular positions."""
    longitudes = np.asarray_chkfinite(longitudes)
    latitudes = np.asarray_chkfinite(latitudes)

    unit_vectors = lonlat_to_unitvec(longitudes, latitudes)
    mean_unit_vector = directional_stats(unit_vectors).mean_direction
    mean_lon, mean_lat = unitvec_to_lonlat(mean_unit_vector)
    return mean_lon[0], mean_lat[0]



from . import axis
from . import process
from . import utilities
from .plot_figure import clustering_result
from .plot_figure import location
from .plot_figure import nearest_neighbor_distances
from .gaia_explorer import GaiaExplorer, ion, ioff
import warnings
warnings.warn(
    "ocelot.plot API will change soon. Most objects will move, change, or be "
    "deprecated.",
    DeprecationWarning,
)



"""A number of functions for pre-processing data before clustering can begin."""

from typing import Optional, Union, Tuple, List

import numpy as np
import pandas as pd
import healpy

from sklearn.preprocessing import RobustScaler, StandardScaler
from astropy.coordinates import SkyCoord
from astropy import units as u
from scipy.optimize import minimize


def cut_dataset(
    data_gaia: pd.DataFrame,
    parameter_cuts: Optional[dict] = None,
    geometric_cuts: Optional[dict] = None,
    return_cut_stars: bool = False,
    reset_index: bool = True,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
    """A function for cutting a dataset based on certain requirements: either on allowed parameter ranges or based on
    geometric cuts (such as selecting a circle from the data.)

    Todo: this function should be able to move clusters that are on an ra boundary.
        E.g. Blanco 1, which always gets messed up!

    Notes:
        - This function will do nothing if parameter_cuts and geometric_cuts are both None (the default settings!).
        - Currently, geometric_cuts is applied first, as this could drastically reduce the dataset size (and speed up
            later applications of parameter_cuts).
        - This could be done a lot more neatly, but is written for speed instead =)

    Args:
        data_gaia (pd.DataFrame): the data to apply the cuts to.
        parameter_cuts (dict, optional): a dictionary of lists or np.ndarrays of allowed [minimum, maximum] parameter
            ranges, in a style like: {"phot_g_mag": [-np.inf, 18]}, where np.inf can be used to ignore limits.
            Default: None
        geometric_cuts (dict, optional): a parameter dictionary for a geometric cut to apply to the dataset. Implemented
            cuts: NONE todo this
            Example: {"type":"great_circle", "sky_position"=[125.429, -16.743]}
        return_cut_stars (bool): whether or not to also return a DataFrame containing *only* the stars that have
            been cut.
            Default: False
        reset_index (bool): whether or not to reset the indexes on the data frames to be returned. May not be intended
            behaviour if you're planning on re-combining the two DataFrames later.
            Default: True

    Returns:
        a cut data_gaia with reset indexing.

    todo this should return the original indices of whatever was dropped, which also need saving so that crossmatching
        can be done later

    """

    if parameter_cuts is not None:
        # Grab the names of the parameters we want to cut based on the dictionary keys.
        list_of_parameters_to_cut = parameter_cuts.keys()

        # Cycle over the parameters to cut and find the stars that satisfy all cuts. It's assumed that data_gaia is
        # oppressively huge, and that dropping & copying the array many times in a row would be inefficient. Instead,
        # we check all parameter matches first, and change data_gaia at the end.
        good_stars = np.ones(data_gaia.shape[0], dtype=bool)

        for a_parameter in list_of_parameters_to_cut:
            a_parameter_cut = parameter_cuts[a_parameter]

            # Check that the range is a list or np.ndarray
            type_to_check = type(a_parameter_cut)
            if type_to_check != list and type_to_check != np.ndarray:
                raise ValueError(
                    "Incorrect type of range found in parameter_cuts! All ranges must be specified as "
                    "Python lists or np.ndarray arrays."
                )

            # Find the matches to the parameter in data_gaia (fyi: it's a lot faster to convert to numpy first. lol)
            # We then only update the good_stars indexes that are already True, so that we don't overwrite existing
            # fails and also only do the minimum amount of calculation.
            column_to_test = data_gaia.loc[good_stars, a_parameter].to_numpy()
            good_stars[good_stars] = np.logical_and(
                column_to_test > a_parameter_cut[0], column_to_test < a_parameter_cut[1]
            )

        # Once we have all the cuts, only *now* do we make data_gaia smaller.
        data_gaia_cut = data_gaia.loc[good_stars, :]

        if reset_index:
            data_gaia_cut = data_gaia_cut.reset_index(drop=True)

        if return_cut_stars:
            data_gaia_dropped_stars = data_gaia.loc[np.invert(good_stars), :]

            if reset_index:
                data_gaia_dropped_stars = data_gaia_dropped_stars.reset_index(drop=True)
        else:
            data_gaia_dropped_stars = None

    else:
        raise ValueError("You must specify parameter_cuts!")

    if geometric_cuts is not None:
        raise NotImplementedError(
            "Sorry, geometric cutting isn't yet implemented here. Poke Emily to do something."
        )

    if return_cut_stars:
        return data_gaia_cut, data_gaia_dropped_stars
    else:
        return data_gaia_cut


def _rotation_function_lat(
    rotation: np.ndarray, coords: SkyCoord, center_coords: SkyCoord
):
    """Returns the least squares value of the 0th/1st and 2nd/3rd latitudes for a given rotation."""
    # The astropy bits
    center_frame = center_coords.skyoffset_frame(rotation=rotation[0] * u.deg)
    coords = coords.transform_to(center_frame)
    lat = coords.lat.value

    # Grab the least squares differences
    return (lat[0] - lat[1]) ** 2 + (lat[2] - lat[3]) ** 2


default_healpy_kwargs = {
    "nest": True,
    "lonlat": True,
    "nside": 32,  # i.e. 2**5
}


def _get_healpix_frame(pixel_id: int, rotate_frame: bool = True, **user_healpy_kwargs):
    """Returns an astropy SkyOffsetFrame given a healpix pixel id denoting the central pixel of the frame.

    Args:
        pixel_id (int): id of the healpix pixel.
        rotate_frame (bool): whether or not to rotate the frame so that the boundaries of the quadrilateral are roughly
            parallel to the latitude, longitude co-ordinate axes.
            Default: True
        **user_healpy_kwargs: kwargs for the healpy methods. You may change the following (shown with their defaults):
            'nest': True
            'lonlat': True
            'nside': 32  # i.e. 2**5

    Returns:
        an astropy SkyOffsetFrame centered on the center of the pixel.

    """
    # Grab arguments for healpy
    healpy_kwargs = default_healpy_kwargs
    healpy_kwargs.update(user_healpy_kwargs)

    # Get the center from the number of sides and the pixel id
    center = healpy.pix2ang(healpy_kwargs["nside"], pixel_id, nest=True, lonlat=True)
    center_coords = SkyCoord(center[0], center[1], frame="icrs", unit="deg")

    # Rotate the frame if requested
    if rotate_frame:
        # Get the four corners of the pixel
        corners_icrs = healpy.vec2ang(
            healpy.boundaries(
                healpy_kwargs["nside"], pixel_id, step=1, nest=healpy_kwargs["nest"]
            ).T,
            lonlat=healpy_kwargs["lonlat"],
        )

        # Make some astropy SkyCoords to handle various bits and bobs and get the rotation angle
        coords = SkyCoord(ra=corners_icrs[0], dec=corners_icrs[1], unit="deg")
        result = minimize(
            _rotation_function_lat, np.asarray([0]), args=(coords, center_coords)
        )

        # Check the result and cast the rotation angle correctly
        if not result.success:
            raise RuntimeError(
                f"failed to converge on an optimum rotation angle for pixel {pixel_id}!"
            )
        rotation_angle = result.x[0] * u.deg

    else:
        rotation_angle = 0 * u.deg

    return center_coords.skyoffset_frame(rotation=rotation_angle)


def recenter_dataset(
    *args,
    center: Optional[Union[tuple, list, np.ndarray]] = None,
    center_type: str = "icrs",
    pixel_id: Optional[int] = None,
    rotate_frame: bool = True,
    proper_motion: bool = True,
    always_return_list: bool = False,
    **user_healpy_kwargs,
) -> Union[pd.DataFrame, List[pd.DataFrame]]:
    """Creates new arbitrary co-ordinate axes centred on centre, allowing for clustering analysis that doesn't get
    affected by distortions. N.B.: currently only able to use ra, dec from data_gaia!

    Args:
        *args (pd.DataFrame): Gaia dataframes to apply the transform to.
        center (list-like): array of length 2 with the ra, dec co-ordinates of the new centre. Must be specified if
            pixel_id is not specified.
            Default: None
        center_type (str): type of frame center is defined in. Must be acceptable by astropy.coordinates.SkyCoord. E.g.
            could be 'icrs' or 'galactic'.
            Default: 'icrs', i.e. center should be (ra, dec).
        pixel_id (int): if working with healpix pixels, this is the id of the central healpix pixel. Must be specified
            if center is not specified.
            Default: None
        rotate_frame (bool): if pixel_id is not None, then you can also ask to have the frame rotated. This will rotate
            the frame so that the boundaries of the quadrilateral pixel are roughly parallel to the latitude, longitude
            co-ordinate axes.
            Default: True
        proper_motion (bool): whether or not to also make transformed proper motions.
            Default: True
        always_return_list (bool): for backwards-compatibility, we always return one DataFrame when only passed one.
            However, setting this to true will make it a one-element list of DataFrames.
            Default: False
        user_healpy_kwargs: kwargs to change for passing to healpy. Will only do something if you're using a pixel_id
            instead of a field center. You may change the following (shown with their defaults):
            'nest': True
            'lonlat': True
            'nside': 32  # i.e. 2**5

    Returns:
        data_gaia, but now with lat, lon, (pmlat, pmlon) keys for the centered data.

    """
    # Deal with if we're using a healpix pixel and hence specify the center as a pixel number
    if center is None and pixel_id is not None:
        center_frame = _get_healpix_frame(
            pixel_id, rotate_frame=rotate_frame, **user_healpy_kwargs
        )

    # Otherwise, deal with the frame being user-specified
    elif center is not None and pixel_id is None:
        center_frame = SkyCoord(
            center[0], center[1], frame=center_type, unit=u.deg
        ).skyoffset_frame()

    else:
        raise ValueError(
            "you must specify one or the other of center and pixel_id. Not both or neither!"
        )

    to_return = []

    for data_gaia in args:
        if proper_motion:
            coords = SkyCoord(
                ra=data_gaia["ra"].to_numpy() << u.deg,
                dec=data_gaia["dec"].to_numpy() << u.deg,
                pm_ra_cosdec=data_gaia["pmra"].to_numpy() << u.mas / u.yr,
                pm_dec=data_gaia["pmdec"].to_numpy() << u.mas / u.yr,
            )
        else:
            coords = SkyCoord(
                ra=data_gaia["ra"].to_numpy() << u.deg,
                dec=data_gaia["dec"].to_numpy() << u.deg,
            )

        # Apply the transform and save it
        coords = coords.transform_to(center_frame)

        data_gaia["lon"] = coords.lon.value
        data_gaia["lat"] = coords.lat.value

        if proper_motion:
            data_gaia["pmlon"] = coords.pm_lon_coslat.value
            data_gaia["pmlat"] = coords.pm_lat.value

        # Correct all values above 180 to simply be negative numbers
        data_gaia["lon"] = np.where(
            data_gaia["lon"] > 180, data_gaia["lon"] - 360, data_gaia["lon"]
        )

        to_return.append(data_gaia)

    if len(to_return) == 1 and always_return_list is False:
        return to_return[0]
    else:
        return to_return


def rescale_dataset(
    data_gaia: pd.DataFrame,
    *args,
    columns_to_rescale: Union[list, tuple] = ("ra", "dec", "pmra", "pmdec", "parallax"),
    column_weights: Union[list, tuple] = (1.0, 1.0, 1.0, 1.0, 1.0),
    scaling_type: str = "robust",
    concatenate: bool = True,
    return_scaler: bool = False,
    **kwargs_for_scaler,
):
    """A wrapper for sklearn's scaling methods that can automatically handle re-scaling data a number of different ways.

    Args:
        data_gaia (pd.DataFrame): the data to apply the scaling to.
        *args (pd.DataFrame): other data frames to apply the same scaling to.
        columns_to_rescale (list, tuple): keys of the columns to re-scale. Should be in the final desired order!
            Default: ('ra', 'dec', 'pmra', 'pmdec', 'parallax')
        column_weights (list, tuple): linear weights to apply to each column after re-scaling, which will reduce their
            impact in nearest-neighbor calculations (see: Liu+19 for an example of this done with parallax)
            Default: (1., 1., 1., 1., 1.)
        scaling_type (str): type of scaler to use, choose from: 'robust' (sklearn.preprocessing.RobustScaler) or
            'standard' (sklearn.preprocessing.StandardScaler). Robust scaling is more appropriate for data with outliers
            Default: 'robust'
        concatenate (bool): whether or not to join all rescaled arrays (assuming multiple args were specified) into one.
            Default: True, so only one np.array is returned.
        return_scaler (bool): whether or not to also return the scaler object for future use.
            Default: False
        **kwargs_for_scaler: additional keyword arguments will be passed to the selected scaler.

    Returns:
        a np.ndarray of shape (n_stars, n_columns) of the re-scaled data, or a list of separate arrays if
            concatenate=False and at least one arg was specified.

    """
    # Make sure that the columns_to_rescale are a list (pandas don't like tuples #fact) and column weights is np.ndarray
    columns_to_rescale = list(columns_to_rescale)
    column_weights = np.asarray(column_weights)

    # Select the scaler to use
    if scaling_type == "robust":
        scaler = RobustScaler(**kwargs_for_scaler)
    elif scaling_type == "standard":
        scaler = StandardScaler(**kwargs_for_scaler)
    else:
        raise ValueError("Selected scaling_type not supported!")

    # Check that all values are finite
    if not np.isfinite(data_gaia[columns_to_rescale].to_numpy()).all():
        raise ValueError(
            "At least one value in data_gaia is not finite! Unable to rescale the data."
        )

    # Apply the scaling & multiply by column weights
    to_return = [
        scaler.fit_transform(data_gaia[columns_to_rescale].to_numpy()) * column_weights
    ]

    # Apply the same scaling scaling to any other data frames, if others have been specified
    for an_extra_cluster in args:
        # Check that all values are finite
        if not np.isfinite(an_extra_cluster[columns_to_rescale].to_numpy()).all():
            raise ValueError(
                "At least one value in args is not finite! Unable to rescale the data."
            )

        # Re-scale!
        to_return.append(
            scaler.transform(an_extra_cluster[columns_to_rescale].to_numpy())
            * column_weights
        )

    # Return a concatenated array
    if concatenate:
        to_return = np.concatenate(to_return, axis=0)

    # Or, just return the one array if no args were specified (slightly ugly code so my API doesn't break lololol)
    elif len(to_return) == 1:
        to_return = to_return[0]

    # Or, return all arrays separately if we aren't concatenating and have multiple arrays (keeps to_return as is)
    # Now we just need to decide on whether or not to also return the scaler.
    if return_scaler:
        return to_return, scaler
    else:
        return to_return



"""A set of tests for use with the pytest module, covering ocelot.isochrone"""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

from pathlib import Path

import numpy as np
import pytest
from astropy.io import ascii

# Path towards the test isochrones
max_label = 7
path_to_test_isochrone = Path('./test_data/isochrones/isochrones.dat')
path_to_test_isochrones = Path('./test_data/isochrones/')
list_of_paths_to_test_isochrones = [Path('./test_data/isochrones/isochrones.dat'),
                                    Path('./test_data/isochrones/isochrones_2.dat')]
path_to_simulated_population = Path('./test_data/simulated_population.dat')


def test_read_cmd_isochrone():
    """Tests the input-output functionality of the isochrone module."""
    my_isochrones = ocelot.isochrone.read_cmd_isochrone(path_to_test_isochrone, max_label=max_label)

    # Check that we've read in the right shape of file
    assert my_isochrones.shape == (2878, 15)

    # Test that the headers were read in correctly
    assert list(my_isochrones.keys()) == \
        ['Zini', 'MH', 'logAge', 'Mini', 'int_IMF', 'Mass', 'logL', 'logTe',
         'logg', 'label', 'mbolmag', 'Gmag', 'G_BPmag', 'G_RPmag', 'G_BP-RP']

    # Check that there aren't any hidden headers in there (CMD 3.3 hides them in some really annoying spots)
    rows_with_another_header = np.where(my_isochrones['Zini'] == '#')[0]
    assert rows_with_another_header.size == 0

    # Check that all the rows have the right max_label
    rows_with_a_bad_label = np.where(my_isochrones['label'] > max_label)[0]
    assert rows_with_a_bad_label.size == 0

    # Check some random values (by extension checking the typing too)
    assert my_isochrones.loc[0, 'Zini'] == 0.0048313
    assert my_isochrones.loc[1000, 'Gmag'] == 7.681
    assert my_isochrones.loc[2877, 'label'] == 3

    return my_isochrones


def test_read_cmd_isochrone_multiple_isochrones():
    """Specifically tests the ability of ocelot.isochrone.read_cmd_isochrone to read multiple files."""
    # Read in isochrones two ways
    isochrones_read_as_directory = ocelot.isochrone.read_cmd_isochrone(path_to_test_isochrones, max_label=max_label)
    isochrones_read_as_list = ocelot.isochrone.read_cmd_isochrone(list_of_paths_to_test_isochrones, max_label=max_label)

    # Check their shapes
    assert isochrones_read_as_directory.shape == (5756, 15)
    assert isochrones_read_as_list.shape == (5756, 15)

    # Check two values near to the end - if indexing went wrong, these will have been scrambled
    assert isochrones_read_as_directory.loc[5000, 'Gmag'] == -0.866
    assert isochrones_read_as_directory.loc[5001, 'Gmag'] == -0.908

    assert isochrones_read_as_list.loc[5000, 'Gmag'] == -0.866
    assert isochrones_read_as_list.loc[5001, 'Gmag'] == -0.908

    return isochrones_read_as_directory


def test_isochrone_interpolation():
    """Tests the isochrone interpolation functionality of ocelot."""
    my_isochrones = ocelot.isochrone.read_cmd_isochrone(path_to_test_isochrone, max_label=max_label)

    # Cut some stars for speed purposes
    stars_to_cut = np.asarray(my_isochrones['MH'] != 0.0).nonzero()[0]
    my_isochrones = my_isochrones.drop(stars_to_cut).reset_index(drop=True)

    isochrones_for_fun_and_profit = ocelot.isochrone.IsochroneInterpolator(my_isochrones,
                                                                           parameters_as_arguments=['logAge'],
                                                                           interpolation_type='LinearND')

    # Test the output, where logAge=6.25 is *not* sampled by the input and should hence give interesting results.
    test_points = np.asarray([[6.], [6.25], [6.5]])
    output_x, output_y = isochrones_for_fun_and_profit(test_points, resolution=100)

    # Check the shapes
    assert output_x.shape == (300,)
    assert output_y.shape == (300,)

    # Check that there aren't any nans or infs
    assert np.count_nonzero(np.isfinite(output_x)) == 300
    assert np.count_nonzero(np.isfinite(output_y)) == 300

    # Check some random numbers
    assert np.allclose(output_x[[0, 100, 200]], [3.551, 3.678, 3.805], rtol=0.0, atol=1e-8)
    assert np.allclose(output_y[[0, 100, 200]], [9.753, 10.0945, 10.436], rtol=0.0, atol=1e-8)

    return [isochrones_for_fun_and_profit, output_x, output_y]


def test_cmd_interpolation():
    """Tests the spline fitting routines that can be applied to a raw CMD from an open cluster."""
    # Grab a test population generated by CMD 3.3
    data_gaia = ascii.read(str(path_to_simulated_population.resolve())).to_pandas()
    data_gaia.columns = ['Z', 'age', 'Mini', 'Mass', 'logL', 'logTe', 'logg', 'label',
                         'mbolmag', 'G', 'G_BP', 'G_RP']

    # Cull the dataset to only include the most massive stars
    bad_stars = np.asarray(data_gaia['Mass'] < 0.5).nonzero()[0]
    data_gaia = data_gaia.drop(bad_stars).reset_index(drop=True)

    # Add some noise
    np.random.seed(42)
    data_gaia[['G_BP', 'G_RP', 'G']] += np.random.normal(loc=0.0, scale=0.03, size=(data_gaia.shape[0], 3))
    data_gaia['G_BP-RP'] = data_gaia['G_BP'] - data_gaia['G_RP']

    # Time to fucking interpolate!!!
    spline_me_up_scotty = ocelot.isochrone.CMDInterpolator(
        parameters=['G_BP-RP', 'G'],
        curve_parameterisation_type='summed',
        data_sorted_on='proximity_to_line',
        filter_input=True,
        filter_window_size=15,
        filter_order=3,
        interp_type='UnivariateSpline',
        interp_smoothing=0.5,
        interp_order=3)
    spline_me_up_scotty.fit(data_gaia, max_repeats=20, print_current_step=False)

    # Time to fucking see the results!!!
    interpolated_cmd_x, interpolated_cmd_y = spline_me_up_scotty(resolution=100)

    # Check that the output values are what we expect
    desired_y = np.asarray([-1.58247131, 2.6352361, 6.23856108, 9.11628436])
    desired_x = np.asarray([1.94389926, 0.75030174, 1.17974861, 2.20645386])

    actual_y = interpolated_cmd_y[[0, 10, 50, 99]]
    actual_x = interpolated_cmd_x[[0, 10, 50, 99]]

    assert np.allclose(desired_y, actual_y, rtol=0.0, atol=1e-8)
    assert np.allclose(desired_x, actual_x, rtol=0.0, atol=1e-8)

    # Todo: If this algorithm gets improved, this quick n' shit unit test should be improved.
    #    (nb at time of writing it didn't look like development would continue any further on this)

    return [data_gaia, interpolated_cmd_x, interpolated_cmd_y]


def test_nn_graph_sort():
    """Tests the nearest neighbour graph sorting routine."""
    # Create some test data
    test_points = 50
    x = np.linspace(0, 2 * np.pi, test_points)
    y = np.sin(x)

    # Randomise the order
    np.random.seed(42)
    random_order = np.random.permutation(test_points)
    x_unsorted = x[random_order]
    y_unsorted = y[random_order]

    # Get the graph to do some graphing!
    sort_args = ocelot.isochrone.interpolate.nn_graph_sort(x_unsorted, y_unsorted)
    x_sorted = x_unsorted[sort_args]
    y_sorted = y_unsorted[sort_args]

    # Useful print code if this test keeps failing:
    # i = 0
    # while i < x_sorted.size:
    #     print(f'{i} ~ {x[i]} {x_sorted[i]} ~ {y[i]} {y_sorted[i]}')
    #     i += 1

    # See if it actually hecking worked:
    # Check that the sorted x array is identical to the original x one
    np.testing.assert_array_max_ulp(x, x_sorted, maxulp=1)

    # Check that the sorted y array is identical to the original y one
    np.testing.assert_array_max_ulp(y, y_sorted, maxulp=1)

    return [x_sorted, y_sorted]


def test_find_nearest_point():
    """Tests a small function that attempts to find the nearest point on a line to a set of data."""
    # Setup some input data
    y = np.repeat(np.arange(5), 2).reshape(5, 2)  # 5 points spaced equally along y=x
    x = (y**2)[:3]  # The squaring makes it so the first and second points match, but then the last matches to y[4]
    desired_result = np.array([0, 1, 4])
    desired_distances = np.array([[0, 0], [0, 0], [0, 0]])

    # Call the function, both with and without raw distances
    result = ocelot.isochrone.interpolate.find_nearest_point(x, y)
    result_distances, distances = ocelot.isochrone.interpolate.find_nearest_point(x, y, return_raw_distances=True)

    # Test that result is right
    np.testing.assert_array_equal(result, desired_result)

    # Test that result is right when distances are returned
    np.testing.assert_array_equal(result_distances, desired_result)

    # Test that the distances are correct
    np.testing.assert_array_equal(distances, desired_distances)

    # Test that giving it an array with the wrong amount of points raises an error
    with pytest.raises(ValueError, match="Input arrays must be two dimensional."):
        ocelot.isochrone.interpolate.find_nearest_point(x[0], y)

    # Test that giving an array with the wrong number of features raises an error
    with pytest.raises(ValueError, match="Number of features mismatch between points_to_match and points_on_line."):
        ocelot.isochrone.interpolate.find_nearest_point(x[:, 0:1], y)


def test_proximity_to_line_sort():
    """Tests the proximity to line sorting (which is mostly just a wrapper to
    ocelot.isochrone.interpolate.find_nearest_point)."""
    # Setup some input data
    y = np.repeat(np.arange(5), 2).reshape(5, 2)  # 5 points spaced equally along y=x
    x = np.array([[0, 0], [1.1, 1.1], [0.9, 0.8], [4.5, 3.8], [4.0, 3.7]])
    desired_result = np.array([0, 2, 1, 4, 3])

    # Run the function
    result = ocelot.isochrone.interpolate.proximity_to_line_sort(x, y)

    # Test that result is right
    np.testing.assert_array_equal(result, desired_result)


# Run tests manually if the file is ran
if __name__ == '__main__':
    import matplotlib.pyplot as plt

    # Run soem tests!!!
    isochrones = test_read_cmd_isochrone()
    isochrones_long = test_read_cmd_isochrone_multiple_isochrones()
    interpolator, interpolator_output_x, interpolator_output_y = test_isochrone_interpolation()
    nn_x_sorted, nn_y_sorted = test_nn_graph_sort()
    test_find_nearest_point()
    test_proximity_to_line_sort()

    # # Plot the results of the nearest neighbour graph sort
    # plt.plot(nn_x_sorted, nn_y_sorted)
    # plt.title('A sorted sine wave that began its life unsorted')
    # plt.show()


    # Some plotting code for cmd spline fitting
    population, cmd_x, cmd_y = test_cmd_interpolation()

    print(population.shape)

    plt.plot(population['G_BP-RP'], population['G'], 'k.')
    plt.plot(cmd_x, cmd_y, 'r-')
    plt.gca().invert_yaxis()
    plt.title('A spline fit to a cmd')
    plt.show()

    """
    # Some plotting code for the isochrone reading & interpolation
    import matplotlib.pyplot as plt

    good_stars = np.asarray(np.logical_and(isochrones['MH'] == 0.0,
                                           np.logical_or(isochrones['logAge'] == 6., isochrones['logAge'] == 6.5)))
    plt.plot(isochrones['G_BP-RP'].iloc[good_stars], isochrones['Gmag'].iloc[good_stars], 'k-', label='original')

    plt.plot(interpolator_output[0], interpolator_output[1], 'r.', label='interpolated')

    plt.gca().invert_yaxis()
    plt.legend()
    plt.title(f'Isochrones: interpolated vs original')
    plt.show()
    """


"""Tools for calculating values based on proper motions and/or velocities."""


def radial_velocity(velocities, with_frame_correction=False):
    # Todo
    pass


def proper_motion(proper_motions, with_frame_correction=False):
    # Todo
    pass



"""Tools for performing likelihood ratio hypothesis tests on cluster candidates."""

import numpy as np
import matplotlib.pyplot as plt

from pathlib import Path

from typing import Union, Optional

from .find import get_field_stars_around_clusters
from .stats import _fit_kth_nn_distribution, _likelihood_ratio_test, _ks_test, _welch_t_test, _mann_whitney_rank_test
from ..cluster.epsilon import kth_nn_distribution
from ..plot.utilities import calculate_alpha

DEFAULT_DIAGNOSTIC_PLOT_SETTINGS = dict(
    dpi=100,
    n_cols=2,
    inches_per_ax=4,
    supertitle_y_coord=1.05,
    file_prefix="",
)


def _plot_points_cropped_on_ax(ax, x_data, y_data, cluster_star_bool_array, field_star_indices):
    """Function for plotting points on a diagnostic axis. Very generalised! Will make sure that the axis is as cropped
    as possible."""
    # Plot the cluster & field stars
    ax.scatter(x_data[cluster_star_bool_array], y_data[cluster_star_bool_array], s=2, c="r", zorder=200)
    ax.scatter(x_data[field_star_indices], y_data[field_star_indices], s=2, c="b", zorder=100)

    # Get the limits that work in this case and then also add other stars
    x_limits, y_limits = ax.get_xlim(), ax.get_ylim()

    good_field_stars = np.logical_and.reduce(
        (x_data > x_limits[0], x_data < x_limits[1], y_data > y_limits[0], y_data < y_limits[1])
    )
    ax.scatter(
        x_data[good_field_stars],
        y_data[good_field_stars],
        s=2,
        c="k",
        zorder=0,
        alpha=calculate_alpha(ax.figure, ax, np.count_nonzero(good_field_stars), 2, scatter_plot=True, max_alpha=0.1),
    )

    # Reset limits (so that the plot is as cropped as possible) and return
    ax.set(xlim=x_limits, ylim=y_limits)
    return ax


def _plot_cluster_probability_distributions(ax, nn_distances, fit_params, cluster_number):
    """Plots the nearest neighbour distributions of the cluster stars and the associated field stars as normed
    probability distributions."""
    # Grab lines of the model fits & normalise them
    x_range = np.linspace(
        0, np.max(np.hstack((nn_distances[0][cluster_number], nn_distances[1][cluster_number]))), num=100
    )

    # Plot some normalised histograms bebe!!
    ax.hist(
        nn_distances[0][cluster_number], bins="auto", histtype="step", color="r", lw=2, label="cluster", density=True
    )
    ax.hist(
        nn_distances[1][cluster_number],
        bins="auto",
        histtype="step",
        color="b",
        lw=2,
        label="local field",
        density=True,
    )
    # ax.hist(true_field_nn_distances, bins='auto', histtype='step', color="k", ls="--", lw=2,
    #            label='true field', density=True)

    # Also plot the fits (if available)
    # For the cluster...
    if fit_params[0] is not None:
        y_cluster = kth_nn_distribution(x_range, *fit_params[0][cluster_number])
        y_cluster /= np.trapz(y_cluster, x=x_range)
        ax.plot(x_range, y_cluster, "r:", lw=2, label="cluster fit")

    # ... and the field!
    if fit_params[1] is not None:
        y_field = kth_nn_distribution(x_range, *fit_params[1][cluster_number])
        y_field /= np.trapz(y_field, x=x_range)
        ax.plot(x_range, y_field, "b:", lw=2, label="l. field fit")

    # Beautify and return
    ax.legend(edgecolor="k")
    ax.set(xlabel="kth nearest neighbour distance", ylabel="probability")
    return ax


def _cluster_significance_diagnostic_plots(
    data_rescaled: np.ndarray,
    labels: np.ndarray,
    nn_distances: Union[tuple, list],
    fit_params: Union[tuple, list],
    cluster_significances: dict,
    field_star_indices: dict,
    test_type: str,
    plot_kwargs: Optional[dict] = None,
    output_dir: Optional[Union[Path, str]] = None,
):
    """Plotter for cluster_significance_test diagnostic plots! Will make a *lot* of plots - one for every cluster!
    Turn on only when you're very stuck. =)

    Args:
        data_rescaled (np.ndarray): array of rescaled data, in shape (n_samples, n_features.)
        labels (np.ndarray): integer label produced by your clustering algorithm of choice for data_rescaled. Noise
            (unclustered stars) should have the label -1. Shape (n_samples,).
        nn_distances (tuple, list): length 2 list-like of nearest neighbour distance dicts for the cluster and the field
            stars respectively.
        fit_params (tuple, list): length 2 list-like of fit parameter dicts for the clusters and field stars
            respectively.
        cluster_significances (dict): a dict of cluster significances to include on the plots.
        field_star_indices (dict): a dict of the indices into data_rescaled of all the field stars selected for each
            cluster.
        test_type (str): the type of test the user requested. We need to know about this in case test_type=='all'!
        plot_kwargs (dict, optional): kwargs to pass to the plotting routine.
            Default: None, which reverts to:
                dpi=100
                ncols=3
                inches_per_ax=4
                supertitle_y_coord=1.05
                file_prefix: ''
        output_dir (str, pathlib.Path, optional): where to save diagnostic plots. If not specified, will revert to
            "./cluster_significance_diagnostic_plots".
            Default: None

    """
    # Setup of input args
    if output_dir is None:
        output_dir = Path("./cluster_significance_diagnostic_plots")
    else:
        output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    if plot_kwargs is not None:
        plot_settings = DEFAULT_DIAGNOSTIC_PLOT_SETTINGS.copy()
        plot_settings.update(plot_kwargs)
    else:
        plot_settings = DEFAULT_DIAGNOSTIC_PLOT_SETTINGS

    # Work out how many plots we'll need and the requisite figure geometry
    required_plots = int(np.ceil(data_rescaled.shape[1] / 2) + 1)
    plot_settings["n_rows"] = int(np.ceil(required_plots / plot_settings["n_cols"]))

    # Loop over every cluster, making a plot!
    for a_cluster in field_star_indices:
        # Matplotlib setup!
        cluster_stars = labels == a_cluster
        fig = plt.figure(
            figsize=(
                plot_settings["inches_per_ax"] * plot_settings["n_cols"],
                plot_settings["inches_per_ax"] * plot_settings["n_rows"],
            ),
            dpi=plot_settings["dpi"],
        )
        axes = []

        # Loop over, plotting the data dimensions
        for i_plot in range(1, required_plots):  # 1 indexing since matplotlib's add_subplot is 1-indexed
            # Grab whichever dimension we're plotting on, being careful if we have an odd number of dimensions to loop
            # back around when we get to the last one
            dim_1 = (i_plot - 1) * 2
            if dim_1 + 1 >= data_rescaled.shape[1]:
                dim_2 = 0
            else:
                dim_2 = dim_1 + 1

            # Plot it!
            axes.append(fig.add_subplot(plot_settings["n_rows"], plot_settings["n_cols"], i_plot))
            axes[-1] = _plot_points_cropped_on_ax(
                axes[-1], data_rescaled[:, dim_1], data_rescaled[:, dim_2], cluster_stars, field_star_indices[a_cluster]
            )
            axes[-1].set(xlabel=f"dim {dim_1}", ylabel=f"dim {dim_2}")

        # Also plot the nearest neighbour probability distributions and model fits
        axes.append(fig.add_subplot(plot_settings["n_rows"], plot_settings["n_cols"], required_plots))
        axes[-1] = _plot_cluster_probability_distributions(axes[-1], nn_distances, fit_params, a_cluster)

        # Process the title, since it can be a bit weird when doing all significances
        # Also, we don't use fig.suptitle (sadly) because that gets cropped by fig.tight_layout, but I can't get rid
        # of fig.tight_layout because it stops things being fucked (I think) when there are lots of plots
        if test_type == "all":
            significance_string = (
                f"lr: {cluster_significances['likelihood'][a_cluster]:.2f}  /  "
                f"t: {cluster_significances['welch_t'][a_cluster]:.2f}  /  "
                f"mw: {cluster_significances['mann_w'][a_cluster]:.2f} \n"
                f"k1+: {cluster_significances['ks_one+'][a_cluster]:.2f}  /  "
                f"k1-: {cluster_significances['ks_one-'][a_cluster]:.2f}  /  "
                f"diff: {cluster_significances['ks_one+'][a_cluster] - cluster_significances['ks_one-'][a_cluster]:.2f}"
                f"\nk2+: {cluster_significances['ks_two+'][a_cluster]:.2f}  /  "
                f"k2-: {cluster_significances['ks_two-'][a_cluster]:.2f}  /  "
                f"diff: {cluster_significances['ks_two+'][a_cluster] - cluster_significances['ks_two-'][a_cluster]:.2f}"
            )
        else:
            significance_string = f"sig: {cluster_significances[a_cluster]}"

        fig.suptitle(
            f"Cluster {a_cluster}\n{significance_string}\n"
            f"cluster/local field members: {np.count_nonzero(cluster_stars)}"
            f" / {len(field_star_indices[a_cluster])}",
            ha="left",
            x=0.10,
            fontsize="medium",
        )

        # Final beautification and output
        fig.subplots_adjust(hspace=0.25, wspace=0.3)
        fig.savefig(
            output_dir / f"{plot_settings['file_prefix']}{a_cluster}_significance.png", dpi=plot_settings["dpi"]
        )

        plt.close(fig)


DEFAULT_KNN_KWARGS = dict(
    overcalculation_factor=3.0,
    min_field_stars=100,
    max_field_stars=500,
    n_jobs=1,
    nn_kwargs=None,
    max_iter=100,
    kd_tree=None,
    cluster_nn_distance_type="internal",
    verbose=False,
)


def cluster_significance_test(
    data_rescaled: np.ndarray,
    labels: np.ndarray,
    min_samples: int = 10,
    knn_kwargs: Optional[dict] = None,
    sampling_resolution: int = 200,
    return_field_star_indices: bool = False,
    make_diagnostic_plots: bool = False,
    plot_kwargs: Optional[dict] = None,
    plot_output_dir: Optional[Union[Path, str]] = None,
    test_type: str = "mann_w",
):
    """Calculates the significance of a cluster by looking at the nearest neighbour distribution of local field
    stars, and performing a hypothesis test to evaluate whether or not cluster stars are more compatible with a cluster
    or the field.

    Args:
        data_rescaled (np.ndarray): array of rescaled data, in shape (n_samples, n_features.)
        labels (np.ndarray): integer label produced by your clustering algorithm of choice for data_rescaled. Noise
            (unclustered stars) should have the label -1. Shape (n_samples,).
        min_samples (int): min_samples used by your clustering algorithm, aka the kth nearest neighbour to
            look at.
            Default: 10
        knn_kwargs (dict, optional): extra kwargs for the kth nearest neighbour process in
            get_field_stars_around_clusters.
            Default: None, which reverts to:
                overcalculation_factor=3.,
                min_field_stars=100,
                max_field_stars=500,
                n_jobs=-1,
                nn_kwargs=None,
                max_iter=100,
                kd_tree=None,
                cluster_nn_distance_type="internal",
        sampling_resolution (int): resolution to sample fitted kth nn models at for likelihood ratio test.
            Default: 200
        return_field_star_indices (bool): whether or not to also return indices of field stars selected by the algorithm
            for your own analysis.
            Default: False
        make_diagnostic_plots (bool): whether or not to output diagnostic plots *for every cluster in labels* showing
            the cluster stars, selected field stars & model fits.
            Default: False
        plot_kwargs (dict, optional): kwargs to pass to the plotting routine.
            Default: None, which reverts to:
                dpi=100
                ncols=3
                inches_per_ax=4
        plot_output_dir (str, pathlib.Path, optional): where to save diagnostic plots. If not specified, will revert to
            "./cluster_significance_diagnostic_plots".
            Default: None

    Returns:
        dict of significances, with cluster labels as keys
        dict of log likelihoods (not normalised by the number of stars in each cluster!), mostly for diagnostics
        (a dict of indices into data_rescaled of selected field stars for each cluster, returned if
            return_field_star_indices is True)
    """

    if knn_kwargs is not None:
        knn_kwargs_to_use = DEFAULT_KNN_KWARGS.copy()
        knn_kwargs_to_use.update(knn_kwargs)
    else:
        knn_kwargs_to_use = DEFAULT_KNN_KWARGS

    # Get all nn distances
    cluster_nn_distances, field_nn_distances, field_star_indices = get_field_stars_around_clusters(
        data_rescaled, labels, min_samples=min_samples, **knn_kwargs_to_use
    )

    # Loop over clusters, doing the tests!!!
    # ALL TESTS
    if test_type == "all":
        # Get fit params
        cluster_fit_params = _fit_kth_nn_distribution(cluster_nn_distances, min_samples, resolution=sampling_resolution)
        field_fit_params = _fit_kth_nn_distribution(field_nn_distances, min_samples, resolution=sampling_resolution)

        # Do the tests
        significances, test_statistics = {}, {}

        significances["likelihood"], test_statistics["likelihood"] = _likelihood_ratio_test(
            cluster_nn_distances, cluster_fit_params, field_fit_params, cdf_resolution=sampling_resolution
        )

        significances["ks_one+"], test_statistics["ks_one+"] = _ks_test(
            cluster_nn_distances, field_fit_params, one_sample=True, cdf_resolution=sampling_resolution
        )
        significances["ks_two+"], test_statistics["ks_two+"] = _ks_test(
            cluster_nn_distances, field_nn_distances, one_sample=False
        )

        significances["ks_one-"], test_statistics["ks_one-"] = _ks_test(
            cluster_nn_distances,
            field_fit_params,
            one_sample=True,
            cdf_resolution=sampling_resolution,
            alternative="less",
        )
        significances["ks_two-"], test_statistics["ks_two-"] = _ks_test(
            cluster_nn_distances, field_nn_distances, one_sample=False, alternative="less"
        )

        significances["welch_t"], test_statistics["welch_t"] = _welch_t_test(cluster_nn_distances, field_nn_distances)

        significances["mann_w"], test_statistics["mann_w"] = _mann_whitney_rank_test(
            cluster_nn_distances, field_nn_distances
        )

    # LIKELIHOOD TEST
    elif test_type == "likelihood":
        cluster_fit_params = _fit_kth_nn_distribution(cluster_nn_distances, min_samples, resolution=sampling_resolution)
        field_fit_params = _fit_kth_nn_distribution(field_nn_distances, min_samples, resolution=sampling_resolution)

        significances, test_statistics = _likelihood_ratio_test(
            cluster_nn_distances, cluster_fit_params, field_fit_params, cdf_resolution=sampling_resolution
        )

    # ONE SAMPLE KS TEST
    elif test_type == "ks_one_sample":
        cluster_fit_params = None
        field_fit_params = _fit_kth_nn_distribution(field_nn_distances, min_samples, resolution=sampling_resolution)
        significances, test_statistics = _ks_test(
            cluster_nn_distances, field_fit_params, one_sample=True, cdf_resolution=sampling_resolution
        )

    # TWO SAMPLE KS TEST
    elif test_type == "ks_two_sample":
        cluster_fit_params, field_fit_params = None, None
        significances, test_statistics = _ks_test(cluster_nn_distances, field_nn_distances, one_sample=True)

    # T TEST  (We do the Welch t-test - not the student's - which doesn't assume equal variance.)
    elif test_type == "t_test":
        cluster_fit_params, field_fit_params = None, None
        significances, test_statistics = _welch_t_test(cluster_nn_distances, field_nn_distances)

    # MANN WHITNEY
    elif test_type == "mann_w":
        cluster_fit_params, field_fit_params = None, None
        significances, test_statistics = _mann_whitney_rank_test(cluster_nn_distances, field_nn_distances)

    # BAD test_type VALUE
    else:
        raise ValueError(
            f"specified test_type {test_type} was not recognised! Please specify one of the following: "
            "'all', 'likelihood', 'ks_one_sample', 'ks_two_sample', 't_test', or 'mann_w'."
        )

    # Diagnostic plot time!
    if make_diagnostic_plots:
        _cluster_significance_diagnostic_plots(
            data_rescaled,
            labels,
            (cluster_nn_distances, field_nn_distances),
            (cluster_fit_params, field_fit_params),
            significances,
            field_star_indices,
            test_type,
            plot_kwargs=plot_kwargs,
            output_dir=plot_output_dir,
        )

    # Return time!
    if return_field_star_indices:
        return significances, test_statistics, field_star_indices, cluster_nn_distances, field_nn_distances
    else:
        return significances, test_statistics



"""A set of functions for nearest neighbor analysis of a clustering field. Especially useful for calculating optimum
DBSCAN/OPTICS epsilon parameters.
"""

import numpy as np

from sklearn.neighbors import NearestNeighbors


def precalculate_nn_distances(
    data: np.ndarray,
    n_neighbors: int = 10,
    n_jobs: int = -1,
    return_sparse_matrix: bool = True,
    return_knn_distance_array: bool = False,
    **kwargs,
):
    """Pre-calculates nearest neighbor (nn) distances for direct plugging into a sklearn clustering algorithm with
    metric=pre-computed. Basically just a wrapper for sklearn.neighbors.NearestNeighbors.

    Args:
        data (np.ndarray): dataset of shape (n_samples, n_features) to calculate nearest neighbor distances for.
        n_neighbors (int): the number of neighbors to calculate upto. Should be as high as the highest min_samples you
            wish to use in your clustering algorithm.
            Default: 10
        n_jobs (int): number of jobs to use to calculate nearest neighbor distances.
            Default: -1 (uses all CPUs)
        return_sparse_matrix (bool): whether or not to return an (n_samples, n_samples) sparse matrix of nearest
            neighbor distances for feeding into a clustering algorithm (e.g. DBSCAN, HDBSCAN*, OPTICS...)
            Default: True
        return_knn_distance_array (bool): whether or not to also return an (n_samples, n_neighbors) distance array,
            useful for determining epsilon for DBSCAN or making nearest neighbor plots.
            Default: False
        **kwargs: any keyword arguments to pass to the sklearn.neighbors.NearestNeighbors constructor.

    Returns:
        a sparse matrix of nearest neighbor distances,
        (a np.ndarray distance array iff return_knn_distance_array is True)

    todo add a way to save output

    """
    # Initialise & fit
    neighbor_calculator = NearestNeighbors(
        n_neighbors=n_neighbors, n_jobs=n_jobs, **kwargs
    )
    neighbor_calculator.fit(data)

    # Return as requested
    if return_knn_distance_array:
        distances, indices = neighbor_calculator.kneighbors()

        if return_sparse_matrix:
            sparse_matrix = neighbor_calculator.kneighbors_graph(mode="distance")
            return sparse_matrix, distances
        else:
            return distances

    elif return_sparse_matrix:
        sparse_matrix = neighbor_calculator.kneighbors_graph(mode="distance")
        return sparse_matrix

    else:
        raise ValueError(
            "Nothing was specified for return. That's probably not intentional!"
        )



"""Tools to find field stars around a list of cluster stars."""

import warnings

import numpy as np
from sklearn.neighbors import NearestNeighbors


def get_field_stars_around_clusters(data_rescaled: np.ndarray, labels, min_samples=10, overcalculation_factor=2.,
                                    min_field_stars=100, max_field_stars=500, n_jobs=1, nn_kwargs=None, max_iter=100,
                                    kd_tree=None, minimum_next_stars_to_check=10, cluster_nn_distance_type="internal",
                                    verbose=False):
    """Gets and returns a cloud of representative field stars around each reported cluster.

    Args:
        data_rescaled (np.ndarray): array of rescaled data, in shape (n_samples, n_features.)
        labels (np.ndarray): labels of the data with shape (n_samples,), in the sklearn format for density-based
            clustering with at least min_samples stars with a noise label of -1.
        min_samples (int): the min_samples nearest neighbour will be returned.
            Default: 10
        overcalculation_factor (float): min_samples*overcalculation_factor nearest neighbors of cluster stars will be
            searched to try and find compatible field stars to evaluate against for the field step.
            Default: 2.
        min_field_stars (int): minimum number of field stars to use in field calculations. If enough aren't found, then
            we'll traverse deeper into the k nearest neighbour graph to find more.
            Default: 100
        max_field_stars (int): as for min_field_stars but a max instead! We won't add more than this many stars.
            Must be less than min_field_stars.
            Default: 500
        n_jobs (int or None): number of jobs to use for calculating nn distances. -1 uses all cores. None uses 1.
            In general, unless the numbers of stars invovled are very large, it can actually be a lot quicker just to
            use one core.
            Default: 1
        nn_kwargs (dict): a dict of kwargs to pass to sklearn.neighbors.NearestNeighbors when running on
            cluster or field stars.
            Default: None
        max_iter (int): maximum number of iterations to run when trying to search for neighbours for a single cluster
            before we raise a RuntimeError.
            Default: 100
        kd_tree (sklearn.neighbors.NearestNeighbors): pre-initialised nearest neighbor searcher (a KD tree unless you
            mad) to look for field stars with. MUST have already been fit to data_rescaled!
            Default: None
        minimum_next_stars_to_check (int): minimum number of stars from a loop that we're checking next time around
            before we automatically increase the overcalculation_factor to ensure that we don't run out of field
            stars to look at.
            Default: 10
        cluster_nn_distance_type (str): type of cluster internal nearest neighbour distance to calculate. If "internal",
            then cluster nn distances will only be made for cluster stars. If "external", these nn distances will also
            include any nearby field stars, which may or may not produce more accurate tails of the cluster nn
            distribution. Depends on what you're going for, though.
            Default: 'internal'
        verbose (bool): whether or not to print to the console about the current cluster we're working on.
            Default: False

    Returns:
        a dict of cluster nn distances
        a dict of field nn distances
        (indices into labels of the field stars used for each field calculation)

    """
    # Process the default nn kwargs
    if nn_kwargs is None:
        nn_kwargs = {}
    nn_kwargs.update(n_neighbors=min_samples, n_jobs=n_jobs)

    # Grab unique labels
    unique_labels, unique_label_counts = np.unique(labels, return_counts=True)

    # Drop any -1 "clusters" (these are in fact noise as reported by e.g. DBSCAN)
    good_clusters = unique_labels != -1
    n_field_stars = unique_label_counts[np.invert(good_clusters)]
    unique_labels = unique_labels[good_clusters]
    unique_label_counts = unique_label_counts[good_clusters]

    # Check that no biscuitry is in progress
    if np.any(unique_label_counts < min_samples):
        raise ValueError(f"one of the reported clusters is smaller than the value of min_samples "
                         f"of {min_samples}! Method will fail.")
    if n_field_stars < len(labels) * 2/3:
        warnings.warn("fewer than 2/3rds of points are field stars! This may be too few to accurately find neighbours "
                      "of cluster points.", RuntimeWarning)
    if min_field_stars > max_field_stars:
        raise ValueError("min_field_stars may not be greater than max_field_stars!")
    if cluster_nn_distance_type == "internal":
        internal_cluster_nn_distances = True
    elif cluster_nn_distance_type == "external":
        internal_cluster_nn_distances = False
    else:
        raise ValueError("selected nearest neighbor distance mode for clusters not recognised! Must be one of "
                         "'internal' or 'external'.")

    # Let's make a KD tree for the entire field and fit it
    if kd_tree is None:
        field_nn_classifier = NearestNeighbors(**nn_kwargs)
        field_nn_classifier.fit(data_rescaled)
    else:
        field_nn_classifier = kd_tree

    # Cycle over each cluster, grabbing stats about its own nearest neighbor distances
    cluster_nn_distances_dict = {}
    field_nn_distances_dict = {}

    field_star_indices_dict = {}

    # Now, let's cycle over everything and find cluster nearest neighbour info
    for a_cluster in unique_labels:
        if verbose:
            print(f"\r    working on cluster {a_cluster}", end="")

        cluster_stars = labels == a_cluster

        # Grab nearest neighbor distances for the cluster & field
        if internal_cluster_nn_distances:
            cluster_nn_distances = _calculate_cluster_nn_distances_internal(data_rescaled, cluster_stars, nn_kwargs)
        else:
            cluster_nn_distances = _calculate_cluster_nn_distances_external(data_rescaled, cluster_stars,
                                                                            field_nn_classifier)

        # We have a lot of error handling for the field to make sure edge cases don't ruin us!
        current_overcalculation_factor = overcalculation_factor
        n_attempts = 0
        while n_attempts < 10:
            try:
                field_star_distances, field_star_indices = _calculate_field_nn_distances(
                    data_rescaled,
                    field_nn_classifier,
                    cluster_stars,
                    max_iter=max_iter,
                    min_field_stars=min_field_stars,
                    max_field_stars=max_field_stars,
                    min_samples=min_samples,
                    minimum_next_stars_to_check=minimum_next_stars_to_check,
                    overcalculation_factor=current_overcalculation_factor)
                n_attempts = 100

            except RuntimeError:
                n_attempts += 1
                if n_attempts >= 10:
                    raise RuntimeError("unable to find neighbors for cluster despite 10 attempts at doing so.")

                warnings.warn(f"failed to find enough neighbors for cluster {a_cluster}! "
                              f"Doubling the overcalculation factor and starting again.", RuntimeWarning)
                current_overcalculation_factor *= 2

        # Save these nearest neighbor distances
        cluster_nn_distances_dict[a_cluster] = cluster_nn_distances[:, min_samples - 1]
        field_nn_distances_dict[a_cluster] = field_star_distances
        field_star_indices_dict[a_cluster] = field_star_indices

    if verbose:
        print("\r    finished nn distance calculations!")

    # Return time!
    return cluster_nn_distances_dict, field_nn_distances_dict, field_star_indices_dict


def _calculate_field_nn_distances(data_rescaled, field_nn_classifier, cluster_stars, max_iter=100, min_field_stars=100,
                                  max_field_stars=500, min_samples=10, minimum_next_stars_to_check=10,
                                  overcalculation_factor=2.):
    """Function for getting the nearest neighbour distribution of field stars around a cluster."""
    # Setup for our loop'in
    current_overcalculation_factor = overcalculation_factor
    n_field_stars = 0
    field_star_distances = []
    field_star_indices = []

    cluster_stars = cluster_stars.nonzero()[0]
    next_stars_to_check = cluster_stars
    already_done_stars = np.asarray([], dtype=int)  # i.e. an array of everything we already calculated a d for

    i = 0
    while n_field_stars < min_field_stars:
        field_n_neighbors = int(np.round(min_samples * current_overcalculation_factor))
        stars_to_check = next_stars_to_check

        # Get some distances!
        field_nn_distances, field_nn_indices = field_nn_classifier.kneighbors(
            data_rescaled[stars_to_check], n_neighbors=field_n_neighbors)

        # Automatically yeet away from the cluster a bit, since most things within min_samples are likely to be
        # cluster stars.
        # if i == 0:
        #     field_nn_indices = field_nn_indices[:, min_samples:]

        # ------ FIND VALID NON-CLUSTER OBJECTS
        # Drop all objects that are connected to cluster stars and any non-unique objects
        # First, test if individual stars are good or bad
        valid_indices = np.isin(field_nn_indices, cluster_stars, invert=True)

        # Then, see if their row has enough non-cluster stars to be unpolluted
        # (we only do this if we aren't on the first step, as on the first step this doesn't matter since
        # our objective is just to move away)
        if i != 0:
            good_rows = np.all(valid_indices[:, :min_samples], axis=1)
            valid_indices = np.logical_and(good_rows.reshape(-1, 1), valid_indices)

        # Remove anything we've done before
        valid_indices[valid_indices] = np.isin(
            field_nn_indices[valid_indices], already_done_stars, invert=True)

        # ------ RECORD VALID STARS
        # If we're on our first run, then we'll loop around again and get some new distances for the
        # now-found field stars. If not, then these distances are for non-cluster stars and we can start
        # to seriously consider stopping the loop
        if i != 0:
            # Get the indices of the valid field stars and get rid of any that would push us over max_field_stars
            remaining_stars = max_field_stars - n_field_stars
            valid_field_stars_at_min_samples = \
                np.unique(valid_indices[:, min_samples - 1].nonzero()[0])[:remaining_stars]

            # Grab the valid field star distances
            valid_field_star_distances = field_nn_distances[
                valid_field_stars_at_min_samples, min_samples - 1]

            # Add them to the running order
            n_field_stars += len(valid_field_star_distances)
            field_star_distances.append(valid_field_star_distances)

            # field_star_indices.append(field_nn_indices[
            #                               valid_field_stars_at_min_samples, min_samples - 1])
            field_star_indices.append(stars_to_check[valid_field_stars_at_min_samples])

        # ------ PREP FOR NEXT LOOP
        # Look for indices of stars we can check next
        next_stars_to_check = np.unique(field_nn_indices[valid_indices.nonzero()])

        # If there aren't enough stars to check next, then use our old list and double the overcalculation factor
        # so that we can hopefully find more next time
        if len(next_stars_to_check) < minimum_next_stars_to_check:
            next_stars_to_check = stars_to_check
            current_overcalculation_factor *= 2

        if current_overcalculation_factor * min_samples > len(data_rescaled):
            raise RuntimeError("unable to find enough neighbors for this cluster, since n_neighbors > n_samples!")

        # Otherwise, we can record these stars as being already done!
        else:
            already_done_stars = np.append(already_done_stars, stars_to_check)

        # Quick check that the while loop isn't the longest thing ever lol
        i += 1
        if i >= max_iter:
            raise RuntimeError(f"unable to traverse the graph of field stars in {max_iter} iterations!")

    # Convert field_star_distances into a 1D array and not a Python list of arrays
    return np.hstack(field_star_distances), np.hstack(field_star_indices)


def _calculate_cluster_nn_distances_internal(data_rescaled, cluster_stars, nn_kwargs):
    """Function for getting the internal nearest neighbour distribution of stars within a cluster.
    (I.e. the distribution but only of cluster stars)
    """
    # Firstly, get the nearest neighbour distances for the cluster alone
    cluster_nn_classifier = NearestNeighbors(**nn_kwargs)
    cluster_nn_classifier.fit(data_rescaled[cluster_stars])

    # And get the nn distances!
    cluster_nn_distances, cluster_nn_indices = cluster_nn_classifier.kneighbors(data_rescaled[cluster_stars])

    return cluster_nn_distances


def _calculate_cluster_nn_distances_external(data_rescaled, cluster_stars, field_nn_classifier):
    """Function for getting the external nearest neighbour distribution of stars within a cluster.
    (I.e. the distribution of both cluster and field stars, may produce more or less accurate distribution tails)
    """
    cluster_nn_distances, cluster_nn_indices = field_nn_classifier.kneighbors(data_rescaled[cluster_stars])

    return cluster_nn_distances



"""A little class for live exploration of a Gaia dataset."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.patches import Rectangle, Circle
import sys
from datetime import datetime

from typing import Union, Optional
from ..cluster.preprocess import cut_dataset
from .utilities import calculate_alpha
from pathlib import Path


DEFAULT_BACKEND = matplotlib.get_backend()


def ion(backend="QT4Agg"):
    """Turns on interactive programming. Essential to use the GaiaExplorer, but called separately to avoid issues!"""
    plt.ion()
    matplotlib.use(backend)


def ioff():
    """Opposite of ion()"""
    plt.ioff()
    matplotlib.use(DEFAULT_BACKEND)


class GaiaExplorer:
    def __init__(self, data_gaia: pd.DataFrame, cluster_location: Union[pd.Series, dict],
                 extra_catalogue_to_plot: Optional[pd.DataFrame] = None,
                 extra_catalogue_cmap: Optional[dict] = None,
                 extra_catalogue_cmap_key: str = "source",
                 magnitude_range: Union[tuple, list, np.ndarray] = ((8, 18), (-1, 5)),
                 initial_guess_factor: float = 2.,
                 adaptive_alpha: bool = True,
                 error_regions_multiplier: Union[int, float] = 0,
                 error_regions_systematic_tol: Union[int, float] = 0,
                 fig_dir: Union[str, Path] = "figs",
                 debug: bool = False):
        """A little class for live exploration of a Gaia dataset.

        Args:
            data_gaia (pd.DataFrame): your standard Gaia dataframe with astrometry & magnitude info.
            cluster_location (pd.Series, dict): information on the location of the cluster to look for.
                Mandatory keys: 'name', 'ra', 'dec', 'pmra', 'pmdec', 'parallax'
                Optional keys for initial guess: 'ang_radius_t', 'pm_dispersion' or 'pmra_std' and 'pmdec_std',
                    'parallax_std'
            magnitude_range (list-like): magnitude range to plot.
                Default: (8, 18)
            initial_guess_factor (float): amount to multiply dispersions by when generating an initial guess zone. A bit
                like a zoom out factor.
                    Default: 2.0  (tends to be a pretty good value)
            adaptive_alpha (bool): whether or not to adaptively change the alpha value while moving around the plot.
                Default: True
            error_regions_multiplier (int or float): size (in sigma) of error regions to add. Only activates when
                greater than 0. If greater than 0, then 'ang_radius_t', 'pmra_error', 'pmdec_error' and 'parallax_error'
                are required keys.
                Default: 0  (off)
            error_regions_systematic_tol (int or float): an extra amount to add on to the error regions for when your
                crossmatching also has a linear systematic tolerance. Only applies to pmra/pmdec/parallax.
                Default: 0
            fig_dir (str or pathlib.Path): directory to save figures to when the user asks to save a figure.
                Default: 'figs'
            debug (bool): if True, print extra stuff to the console.
                Default: False
        """
        # Initial setup of the data
        self.data = data_gaia
        self.data['bp_rp'] = self.data['phot_bp_mean_mag'] - self.data['phot_rp_mean_mag']
        self.fig_dir = Path(fig_dir)

        # Check we have the minimum amount of information
        required_keys = ("name", "ra", "dec", "pmra", "pmdec", "parallax")
        if not np.all(np.isin(required_keys, tuple(cluster_location.keys()))):
            raise ValueError("initial cluster location must contain all 5 astrometric dimensions and a name!")

        # Setup error region highlighting
        self.error_regions_multiplier = error_regions_multiplier
        self.error_regions_systematic_tol = error_regions_systematic_tol
        if self.error_regions_multiplier > 0:
            required_keys = ("ang_radius_t", "pmra_error", "pmdec_error", "parallax_error")
            if not np.all(np.isin(required_keys, tuple(cluster_location.keys()))):
                raise ValueError("initial cluster location must contain the following keys to add error regions: "
                                 "'ang_radius_t', 'pmra_error', 'pmdec_error' and 'parallax_error'.")

        # Initial setup of the catalogue too
        self.literature_data = extra_catalogue_to_plot
        if self.literature_data is not None:
            # Check that we have something to use for the cmap key
            if extra_catalogue_cmap_key not in self.literature_data.keys():
                raise ValueError("Column not found: to decide how to colour code literature objects, a column with name"
                                 " 'extra_catalogue_cmap_key' must be in extra_catalogue_to_plot!")

            # Generate a cmap if one wasn't provided
            self.literature_data_cmap_key = extra_catalogue_cmap_key
            if extra_catalogue_cmap is None:
                unique_cmap_keys = np.unique(self.literature_data[self.literature_data_cmap_key])
                cmap = plt.get_cmap("tab10")(np.arange(len(unique_cmap_keys)))
                self.literature_data_cmap = {k: v for (k, v) in zip(unique_cmap_keys, cmap)}
            else:
                self.literature_data_cmap = extra_catalogue_cmap

        # Guess initial limit information if it isn't given
        # Firstly, guess the angular radius: assume 20pc if none given, or a minimum of 0.1 degrees
        if "ang_radius_t" not in cluster_location.keys():
            cluster_location["ang_radius_t"] = np.clip(
                np.arctan(20 / (1000 / cluster_location['parallax'])) * 180 / np.pi, 0.1, np.inf)

        # Simple pm dispersion guess of 2 mas/yr if none given
        if "pm_dispersion" not in cluster_location.keys():
            if np.all(np.isin(("pmra_std", "pmdec_std"), cluster_location.keys())):
                cluster_location["pm_dispersion"] = np.sqrt(
                    cluster_location["pmra_std"]**2 + cluster_location["pmdec_std"]**2)
            else:
                cluster_location["pm_dispersion"] = 2.

        # Simple 0.5 mas parallax guess if none given
        if "parallax_std" not in cluster_location.keys():
            cluster_location["parallax_std"] = 0.5

        # Finally, write this initial limit guess to the class
        guess_array = np.asarray([-initial_guess_factor, initial_guess_factor])
        self.cluster_location = cluster_location
        self.current_cuts = dict(
            ra=guess_array * cluster_location["ang_radius_t"] + cluster_location['ra'],
            dec=guess_array * cluster_location["ang_radius_t"] + cluster_location['dec'],
            pmra=guess_array * cluster_location["pm_dispersion"] + cluster_location['pmra'],
            pmdec=guess_array * cluster_location["pm_dispersion"] + cluster_location['pmdec'],
            parallax=guess_array * cluster_location["parallax_std"] + cluster_location['parallax'],
        )
        self.magnitude_cuts = dict(
            phot_g_mean_mag=np.sort(magnitude_range[0]),
            bp_rp=np.sort(magnitude_range[1])
        )

        # We'll also want to consider some move and zoom speeds. The latter is just set to two (we'll double or halve
        # the zoom every time) but the former uses the dispersions etc to make an educated first guess.
        self.move_speed = np.asarray([
            [cluster_location["ang_radius_t"], cluster_location["ang_radius_t"]],  # ra / dec
            [cluster_location["pm_dispersion"], cluster_location["pm_dispersion"]],  # pmra / pmdec
            [cluster_location["ang_radius_t"], cluster_location["parallax_std"]],  # ra / parallax
        ])

        self.zoom_speed = [2, 2, 2]

        # alpha-value related info
        self.adaptive_alpha = adaptive_alpha
        self.alpha_multiplier = 1.
        self.current_alpha = 1.
        self.first_update_data_call = True

        # Other info we might need
        self.fig, self.ax, self.main_lines = None, None, None
        self.cluster_name = cluster_location["name"]
        self.current_axis = 0
        self.axis_labels = (
            ("ra", "dec"),
            ("pmra", "pmdec"),
            ("ra", "parallax"),
            ("BP - RP", "G"),
        )

        self.debug = debug

    def _generate_figure(self):
        """Generates a figure for the class to use with four different panels."""
        fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(8, 8))
        ax = ax.flatten()

        # Assign labels to everyone
        for an_ax, labels in zip(ax, self.axis_labels):
            an_ax.set(xlabel=labels[0], ylabel=labels[1])

        # Magnitudes are a special case
        ax[3].set(xlim=self.magnitude_cuts['bp_rp'], ylim=self.magnitude_cuts['phot_g_mean_mag'])
        ax[3].invert_yaxis()

        # Plot initial objects
        self.main_lines = []
        for an_ax in ax:
            self.main_lines.append(an_ax.plot([], [], 'ko', ms=3)[0])

        # Save to the class and add the initial data
        self.fig, self.ax = fig, ax
        self._add_catalogue_clusters()
        self._add_error_regions()
        self._update_data()
        self._update_active_axis(0)
        self._initialise_matplotlib_connections()
        plt.show()

    def _refresh_figure(self):
        """Refreshes the figure so that any change is displayed."""
        self.fig.canvas.draw()
        # self.fig.canvas.flush_events()

    def _add_catalogue_clusters(self):
        """Call for updating the figure with locations of clusters as marked by other catalogues."""
        if self.literature_data is not None:
            if len(self.literature_data) > 0:
                # Cycle over every literature object
                for index_literature, a_row in self.literature_data.iterrows():
                    color = self.literature_data_cmap[a_row[self.literature_data_cmap_key]]

                    # Add it to every axis!
                    for i_ax in range(3):
                        x_value, y_value = a_row[self.axis_labels[i_ax][0]], a_row[self.axis_labels[i_ax][1]]
                        name = " ".join(a_row['name'].rsplit("_"))
                        self.ax[i_ax].plot(x_value, y_value, 'x', ms=5, color=color)  # , mec='w', mew=0.5)
                        self.ax[i_ax].text(
                            x_value, y_value, name, va='bottom', ha='center', color=color, fontsize=8, clip_on=True)
                        # .set_path_effects([patheffects.withStroke(linewidth=0.5, foreground='w')])

            self._refresh_figure()

    def _add_error_regions(self):
        """Adds error regions to the figure if requested by the user."""
        if self.error_regions_multiplier > 0:
            # Firstly, let's try and grab a colour for our circle if it's in the literature_data list and hence has
            # a consistent catalogue colour we could use
            if self.literature_data is not None:
                if np.all(np.isin(self.cluster_location['name'], self.literature_data['name'])):
                    first_match = np.asarray(self.literature_data['name'].to_numpy() == self.cluster_location['name']
                                             ).nonzero()[0][0]
                    color = self.literature_data_cmap[
                        self.literature_data.loc[first_match, self.literature_data_cmap_key]]
                else:
                    color = 'r'

            # Otherwise, just use red!
            else:
                color = 'r'

            # Add them to the plots!
            style = dict(fc='None', ec=color)

            # ra/dec
            location = (self.cluster_location['ra'], self.cluster_location['dec'])
            self.ax[0].add_patch(Circle(location, radius=self.cluster_location['ang_radius_t'], **style))

            # pmra/pmdec
            width = self.cluster_location['pmra_error'] * 2 * self.error_regions_multiplier \
                + 2 * self.error_regions_systematic_tol
            height = self.cluster_location['pmdec_error'] * 2 * self.error_regions_multiplier \
                + 2 * self.error_regions_systematic_tol
            location = (self.cluster_location['pmra'] - width / 2, self.cluster_location['pmdec'] - height / 2)
            self.ax[1].add_patch(Rectangle(location, width, height, **style))

            # ra/parallax
            width = self.cluster_location['ang_radius_t']
            height = self.cluster_location['parallax_error'] * 2 * self.error_regions_multiplier \
                + 2 * self.error_regions_systematic_tol
            location = (self.cluster_location['ra'] - width / 2, self.cluster_location['parallax'] - height / 2)
            self.ax[2].add_patch(Rectangle(location, width, height, **style))

    def _update_data(self):
        """Call for updating the figure with new cuts."""
        # Get the new data
        new_data = cut_dataset(self.data, parameter_cuts=self.current_cuts)

        # Get an alpha estimate. We just do this for the first one cos it'll stay the same
        if self.adaptive_alpha or self.first_update_data_call:
            self.current_alpha = calculate_alpha(self.fig, self.ax[0], len(new_data), 3)
            self.first_update_data_call = False

        # Apply our current alpha value to the data
        self.alpha_to_use = np.clip(self.current_alpha * self.alpha_multiplier, 0.01, 1.0)
        for i in range(4):
            self.main_lines[i].set_alpha(self.alpha_to_use)

        # Update all of the lines with the new stuff
        self.main_lines[0].set_data(new_data['ra'], new_data['dec'])
        self.main_lines[1].set_data(new_data['pmra'], new_data['pmdec'])
        self.main_lines[2].set_data(new_data['ra'], new_data['parallax'])
        self.main_lines[3].set_data(new_data['bp_rp'], new_data['phot_g_mean_mag'])

        # Change the limits toooo
        for i in range(3):
            self.ax[i].set(xlim=self.current_cuts[self.axis_labels[i][0]],
                           ylim=self.current_cuts[self.axis_labels[i][1]])

        # Add a title cos we cool
        self.fig.suptitle(f"The region around {self.cluster_name}\nstars: {len(new_data)}", x=0.1, y=0.98, ha='left')

        self._refresh_figure()

    def _update_active_axis(self, new_axis: int):
        """Updates which axis is the active one, both internally and with visual cues."""
        # Firstly, reset the colour of the old one
        an_ax = self.ax[self.current_axis]
        for a_spine in an_ax.spines:
            an_ax.spines[a_spine].set_color('k')

        # Now, change the active axis
        self.current_axis = new_axis

        # ... and change the colour of the new one
        an_ax = self.ax[self.current_axis]
        for a_spine in an_ax.spines:
            an_ax.spines[a_spine].set_color('r')

        self._refresh_figure()

    def _update_move_speed(self, go_faster: bool):
        """Changes the speed at which we move around the current axis."""
        if go_faster:
            self.move_speed[self.current_axis] *= 2
        else:
            self.move_speed[self.current_axis] /= 2

    def _change_zoom(self, go_in: bool):
        """Changes the zoom level on the current axis."""
        # Firstly, let's get the current difference between x and y values
        x_cuts = self.current_cuts[self.axis_labels[self.current_axis][0]]
        y_cuts = self.current_cuts[self.axis_labels[self.current_axis][1]]

        differences = np.asarray([np.diff(x_cuts)[0], np.diff(y_cuts)[0]])
        midpoints = np.asarray([np.mean(x_cuts), np.mean(y_cuts)])

        # We can compute the new difference we need
        if go_in:
            new_differences = differences / self.zoom_speed[self.current_axis] / 2
            self._update_move_speed(False)
        else:
            new_differences = differences * self.zoom_speed[self.current_axis] / 2
            self._update_move_speed(True)

        # Hard coded special case that we won't change ra values if we're editing the ra/parallax plot
        if self.current_axis == 2:
            i_start = 1
        else:
            i_start = 0

        # Finally, we can write this to the limits!
        for i in range(i_start, 2):
            self.current_cuts[self.axis_labels[self.current_axis][i]] = (
                midpoints[i] + np.asarray([-new_differences[i], new_differences[i]]))

        # ... and run the data updater
        self._update_data()

    def _change_location(self, key: str):
        """Changes the midpoint of the current axis, i.e. the area we're looking at."""
        # Firstly, we need to process the direction
        # w or s are up/down, i.e. y axis
        if key in ("up", "down"):
            current_axis_name = self.axis_labels[self.current_axis][1]
            is_y_axis = 1
        else:
            current_axis_name = self.axis_labels[self.current_axis][0]
            is_y_axis = 0

        # w or d mean there's an increase
        if key in ("up", "right"):
            direction = 1
        else:
            direction = -1

        # Now, we can add this to the limits! Easy pweezy. And update the data too
        self.current_cuts[current_axis_name] += direction * self.move_speed[self.current_axis, is_y_axis]
        self._update_data()

    def _change_alpha(self, increase: bool):
        """Changes the alpha value that we plot our data points with."""
        if increase:
            self.alpha_multiplier *= 1.5
        else:
            self.alpha_multiplier /= 1.5

        self._update_data()

    def _save_figure(self):
        """Saves the current view of the figure!"""
        # Make sure the folder exists and save to it!
        self.fig_dir.mkdir(parents=True, exist_ok=True)
        now = datetime.now()
        date_string = f"{now.year}-{now.month:0>2}-{now.day:0>2}--{now.hour:0>2}-{now.minute:0>2}-{now.second:0>2}"
        fig_name = f"{self.cluster_name}_{date_string}.png"
        self.fig.savefig(self.fig_dir / fig_name, bbox_inches="tight")
        print(f"  saved figure {fig_name}")
        sys.stdout.flush()

    def _close_figure(self):
        plt.close(self.fig)

    def _initialise_matplotlib_connections(self):
        """Starts matplotlib connections that mean we call self._handle_keypress when a key is pressed on the window"""
        self.fig.canvas.mpl_connect('key_press_event', self._handle_keypress)

    def _handle_keypress(self, event):
        """Core function for updates of the figure element"""
        if self.debug:
            print("Key pressed:", event.key)
            sys.stdout.flush()

        # Change current axis being edited
        if event.key in ("1", "2", "3"):
            self._update_active_axis(int(event.key) - 1)

        # Change current axis move or zoom speed
        elif event.key in ("*", "/"):
            self._update_move_speed(event.key == "*")

        # Change current axis zoom level
        elif event.key in ("+", "-"):
            self._change_zoom(event.key == "+")

        # Change current axis center location
        elif event.key in ("up", "right", "down", "left"):
            self._change_location(event.key)

        # Change current alpha multiplier
        elif event.key in ("pageup", "pagedown"):
            self._change_alpha(event.key == "pageup")

        # Save the current figure
        elif event.key == 'end':
            self._save_figure()

        # Quit the figure
        elif event.key == "q":
            self._close_figure()

        # Or alternatively, key not supported
        else:
            if self.debug:
                print("  key not recognised")
                sys.stdout.flush()

    def __call__(self):
        """Plot a Gaia data chunk and let the user move around!"""
        self._generate_figure()



"""A set of tests for ocelot.calculate."""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

import pickle
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

path_to_blanco_1_cluster = Path('./test_data/blanco_1_gaia_dr2_gmag_18_cut_cluster_only.pickle')


def test__weighted_standard_deviation():
    """Tests ocelot.calculate.calculate._weighted_standard_deviation, a function for computing weighted standard
    deviations."""
    # Create some simple test data which clearly has standard deviation ~1, other than ten cheeky points
    np.random.seed(42)
    length = 10000
    normal_points = np.random.normal(loc=0.0, scale=1.0, size=length)

    not_so_normal_points = np.empty((length + 1000,), dtype=float)
    not_so_normal_points[:length] = normal_points
    not_so_normal_points[length:] = 10000.

    weights = np.ones(not_so_normal_points.shape)
    weights[length:] = 0.

    # Do our calculations
    weighted_std_normal_points = ocelot.calculate.calculate._weighted_standard_deviation(normal_points, None)
    weighted_std_abnormal_points = ocelot.calculate.calculate._weighted_standard_deviation(not_so_normal_points,
                                                                                           weights)
    std_normal_points = np.std(normal_points)

    # Check that my function and the actual standard deviation are about the same
    assert np.allclose(weighted_std_normal_points, std_normal_points, rtol=0.0, atol=1e-3)

    # Check that the weighting correctly ignores the dodgy points
    assert np.allclose(weighted_std_abnormal_points, std_normal_points, rtol=0.0, atol=1e-3)


def parameter_return_test_function(parameters, target, rtol=0.0, atol=1e-8):
    """Accepts a parameter dict and a target dict, like those returned by ocelot.calculate.

    Tests that:
        1. The number of keys is the same
        2. The keys are the same
        3. The values are the same

    """
    # Check that the keys are right (indicative of a code change)
    keys_parameters = list(parameters.keys())
    keys_target = list(target.keys())
    assert len(keys_parameters) == len(keys_target)

    for i in range(len(keys_parameters)):
        # Check the keys are the same, in the same location
        assert keys_parameters[i] == keys_target[i]

        # Check the value (done differently depending on if it is or isn't a string)
        if isinstance(target[keys_parameters[i]], str):
            assert parameters[keys_parameters[i]] == target[keys_parameters[i]]
        else:
            assert np.allclose(parameters[keys_parameters[i]], target[keys_parameters[i]],
                               rtol=rtol, atol=atol, equal_nan=True)


def test_mean_distance():
    """Tests ocelot.calculate.mean_distance"""
    # Read in Blanco 1
    with open(path_to_blanco_1_cluster, 'rb') as handle:
        data_cluster = pickle.load(handle)

    # Make the ra values non-corrected
    data_cluster['ra'] = np.where(data_cluster['ra'] < 0, data_cluster['ra'] + 360, data_cluster['ra'])

    # Calculate distances
    parameters = ocelot.calculate.mean_distance(data_cluster)

    target_dict = {
        'parallax': 4.124794980897457,
        'parallax_std': 0.3915133864674593,
        'parallax_error': 0.024187777793597237,
        'inverse_parallax': 242.4362918959972,
        'inverse_parallax_l68': 241.02293457480792,
        'inverse_parallax_u68': 243.86632283215837,
        'distance': 244.3519932235079,
        'distance_std': 36.51808584564543,
        'distance_error': 2.25609487801101
    }

    parameter_return_test_function(parameters, target_dict)


def test_mean_proper_motion():
    """Tests ocelot.calculate.mean_proper_motion"""
    # Read in Blanco 1
    with open(path_to_blanco_1_cluster, 'rb') as handle:
        data_cluster = pickle.load(handle)

    # Make the ra values non-corrected
    data_cluster['ra'] = np.where(data_cluster['ra'] < 0, data_cluster['ra'] + 360, data_cluster['ra'])

    # Calculate the mean proper motion
    parameters = ocelot.calculate.mean_proper_motion(data_cluster)

    target_dict = {
        'pmra': 18.691800985496148,
        'pmra_std': 0.7782083402189158,
        'pmra_error': 0.04807787181985345,
        'pmdec': 2.6059267710228995,
        'pmdec_std': 0.7849144024486323,
        'pmdec_error': 0.0484921737280103,
        'pm_dispersion': 1.1053048629032505}

    parameter_return_test_function(parameters, target_dict)


def test_mean_radius():
    """Tests ocelot.calculate.mean_radius"""
    # Read in Blanco 1
    with open(path_to_blanco_1_cluster, 'rb') as handle:
        data_cluster = pickle.load(handle)

    # Make the ra values non-corrected
    data_cluster['ra'] = np.where(data_cluster['ra'] < 0, data_cluster['ra'] + 360, data_cluster['ra'])

    # Calculate the mean proper motion
    parameters = ocelot.calculate.mean_radius(data_cluster)

    target_dict = {
        'ra': 0.9211424874405434,
        'ra_std': 0.42919239850712043,
        'ra_error': 0.026515594931398594,
        'dec': -29.968388740371715,
        'dec_std': 0.37079687952364343,
        'dec_error': 0.022907907720347058,
        'ang_dispersion': 0.5671828988966136,
        'ang_radius_50': 0.4704416634430194,
        'ang_radius_50_error': np.nan,
        'ang_radius_c': np.nan,
        'ang_radius_c_error': np.nan,
        'ang_radius_t': 0.9014932764089681,
        'ang_radius_t_error': np.nan,
        'radius_50': 1.9906299639509861,
        'radius_50_error': np.nan,
        'radius_c': np.nan,
        'radius_c_error': np.nan,
        'radius_t': 3.814813688665708,
        'radius_t_error': np.nan}

    parameter_return_test_function(parameters, target_dict)


def test_mean_internal_velocity_dispersion():
    """Tests ocelot.calculate.internal_velocity_dispersion"""
    # Read in Blanco 1
    with open(path_to_blanco_1_cluster, 'rb') as handle:
        data_cluster = pickle.load(handle)

    # Make the ra values non-corrected
    data_cluster['ra'] = np.where(data_cluster['ra'] < 0, data_cluster['ra'] + 360, data_cluster['ra'])

    # Calculate the mean proper motion
    parameters = ocelot.calculate.mean_internal_velocity_dispersion(data_cluster)

    target_dict = {
        'v_internal_tangential': 931.3613254058669,
        'v_internal_tangential_error': np.nan}

    parameter_return_test_function(parameters, target_dict)


def test_all_statistics():
    """Tests ocelot.calculate.all_statistics"""
    # Read in Blanco 1
    with open(path_to_blanco_1_cluster, 'rb') as handle:
        data_cluster = pickle.load(handle)

    # Make the ra values non-corrected
    data_cluster['ra'] = np.where(data_cluster['ra'] < 0, data_cluster['ra'] + 360, data_cluster['ra'])

    # Calculate the mean proper motion
    parameters = ocelot.calculate.all_statistics(data_cluster, mode="mean")

    target_dict = {'n_stars': 262, 'ra': 0.9211424874405434, 'ra_std': 0.42919239850712043,
                   'ra_error': 0.026515594931398594, 'dec': -29.968388740371715, 'dec_std': 0.37079687952364343,
                   'dec_error': 0.022907907720347058, 'ang_dispersion': 0.5671828988966136,
                   'ang_radius_50': 0.4704416634430194, 'ang_radius_50_error': np.nan, 'ang_radius_c': np.nan,
                   'ang_radius_c_error': np.nan, 'ang_radius_t': 0.9014932764089681, 'ang_radius_t_error': np.nan,
                   'radius_50': 1.9906299639509861, 'radius_50_error': np.nan, 'radius_c': np.nan, 'radius_c_error': np.nan,
                   'radius_t': 3.814813688665708, 'radius_t_error': np.nan, 'parallax': 4.124794980897457,
                   'parallax_std': 0.3915133864674593, 'parallax_error': 0.024187777793597237,
                   'inverse_parallax': 242.4362918959972, 'inverse_parallax_l68': 241.02293457480792,
                   'inverse_parallax_u68': 243.86632283215837, 'distance': 244.3519932235079,
                   'distance_std': 36.51808584564543, 'distance_error': 2.25609487801101, 'pmra': 18.691800985496148,
                   'pmra_std': 0.7782083402189158, 'pmra_error': 0.04807787181985345, 'pmdec': 2.6059267710228995,
                   'pmdec_std': 0.7849144024486323, 'pmdec_error': 0.0484921737280103,
                   'pm_dispersion': 1.1053048629032505, 'v_internal_tangential': 931.3613254058669,
                   'v_internal_tangential_error': np.nan, 'parameter_inference_mode': 'mean'}

    parameter_return_test_function(parameters, target_dict)

    return parameters


def test_points_on_sphere(show_diagnostic_histograms=False):
    """A function for testing the random spherical co-ordinates generator."""
    np.random.seed(42)

    # Calls and a basic diagnostic plot if requested
    theta_rad, phi_rad = ocelot.calculate.random.points_on_sphere(3000, radians=True, phi_symmetric=False)
    theta_deg, phi_deg = ocelot.calculate.random.points_on_sphere(3000, radians=False, phi_symmetric=True)

    if show_diagnostic_histograms:
        fig, ax = plt.subplots(1, 2, figsize=(8, 4))
        ax[0].hist(theta_deg, bins='auto')
        ax[0].set_title("distribution of theta in degrees")
        ax[1].hist(phi_deg, bins='auto')
        ax[1].set_title("distribution of phi in degrees")
        fig.show()
        plt.close('all')

    # Check radian input_mode, asymmetric
    assert np.all(np.logical_and(theta_rad >= 0, theta_rad < 2 * np.pi))
    assert np.all(np.logical_and(phi_rad >= 0, phi_rad <= np.pi))

    # Check degree input_mode, symmetric
    assert np.all(np.logical_and(theta_deg >= 0, theta_deg < 360))
    assert np.all(np.logical_and(phi_deg >= -90, phi_deg <= 90))

    # Also check that phi appears correctly distributed as it's the hard one here
    assert np.allclose(np.std(phi_deg), 39.22, atol=0.5)


if __name__ == "__main__":
    test__weighted_standard_deviation()
    test_mean_distance()
    test_mean_proper_motion()
    test_mean_radius()
    test_mean_internal_velocity_dispersion()
    par = test_all_statistics()
    test_points_on_sphere(show_diagnostic_histograms=True)



"""A set of tests for use with the pytest module, covering ocelot.isochrone"""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

import numpy as np
import pickle
import pytest
from pathlib import Path

path_blanco_1 = Path("./test_data/blanco_1_gaia_dr2_gmag_18_cut.pickle")
path_blanco_1_labels = Path("./test_data/blanco_1_gaia_dr2_gmag_18_cut_LABELS.pickle")


def test_cluster_significance_test():
    """Tests the cluster significance test from ocelot.verify.significance.py. Should be fun!"""
    # Grab the data
    with open(path_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)
    with open(path_blanco_1_labels, 'rb') as handle:
        labels = pickle.load(handle)

    # Re-scale and generally pre-process it!
    data_gaia = ocelot.cluster.recenter_dataset(data_gaia, center=(data_gaia['ra'].mean(), data_gaia['dec'].mean()))
    data_rescaled = ocelot.cluster.rescale_dataset(
        data_gaia, columns_to_rescale=('lon', 'lat', 'pmlon', 'pmlat', 'parallax'))

    # Insert two fake clusters: the first is a selection of non-overdense stars and the second is just some random stars
    # in the data that are totally unclustered!
    np.random.seed(42)
    unclustered_stars = labels == -1
    cluster_1 = np.logical_and.reduce((unclustered_stars,
                                       data_gaia['lon'] > 0, data_gaia['lon'] < 1,
                                       data_gaia['lat'] > 0, data_gaia['lat'] < 1,
                                       data_gaia['pmlon'] > -5, data_gaia['pmlon'] < 5,
                                       data_gaia['pmlat'] > -5, data_gaia['pmlat'] < 5,
                                       data_gaia['parallax'] > 0.25, data_gaia['parallax'] < 3,))
    cluster_2 = np.logical_and.reduce((unclustered_stars, np.invert(cluster_1),
                                       np.random.uniform(size=len(labels)) > 0.975))
    labels[cluster_1] = 1
    labels[cluster_2] = 2

    # Check that the setup worked
    unique_labels, counts = np.unique(labels, return_counts=True)
    assert np.all(counts == np.asarray([14111, 262, 65, 347]))

    # Get us some significances!
    significances, log_likelihoods = ocelot.verify.cluster_significance_test(
        data_rescaled,
        labels,
        min_samples=10,
        make_diagnostic_plots=True,
        plot_output_dir=Path("./test_results/cluster_significance_test3"),
        test_type="all"
    )


if __name__ == "__main__":
    test_cluster_significance_test()




"""Statistical tools (like tests and kth nn distribution tools) for use in verification of OCs."""

import numpy as np
import warnings
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from scipy.stats import norm, chi2, kstest, ttest_ind_from_stats, mannwhitneyu

from ocelot.cluster.epsilon import kth_nn_distribution, constrained_a


class KthNNDistributionCDF:
    def __init__(self, a, d, k, grid_size=200, limits=(0, 1)):
        """Precalculates a grid of kth nn distribution CDF values for a given a, d and k."""
        # Precalculate cumulatively summed values!
        x_range = np.linspace(*limits, num=grid_size)
        y_range = np.cumsum(kth_nn_distribution(x_range, a, d, k))

        # Normalise these values so that the max val is just 1!
        y_range /= np.max(y_range)

        # Interpolate this and set it to __call__!
        self.cdf_function = interp1d(x_range, y_range, kind="linear", fill_value=(0., 1.), bounds_error=False)

    def __call__(self, x):
        return self.cdf_function(x)


class KthNNDistributionPDF:
    def __init__(self, a, d, k, grid_size=200, limits=(0, 1)):
        """Precalculates a grid of kth nn distribution PDF values for a given a, d and k."""
        # Precalculate cumulatively summed values!
        x_range = np.linspace(*limits, num=grid_size)
        y_range = kth_nn_distribution(x_range, a, d, k)

        # Normalise these values so that the max val is just 1!
        y_range /= np.trapz(y_range, x=x_range)

        # Interpolate this and set it to __call__!
        self.pdf_function = interp1d(x_range, y_range, kind="linear", fill_value=(0., 0.), bounds_error=False)

    def __call__(self, x):
        return self.pdf_function(x)


class _CurveToFit:
    def __init__(self, k):
        self.k = k

    def __call__(self, x_range, *params):
        """A version of the Chandrasekhar1943 curve for use with scipy's curve_fit function.

        Args:
            x_range: should *already* be cumulatively summed and sorted wrt point number!
            params: (a, dimension)

        Returns:
            y_range of points evaluated at x_range values!

        """
        y_func = np.cumsum(kth_nn_distribution(x_range, params[0], params[1], self.k))
        y_func /= np.trapz(y_func, x=x_range)

        return y_func


def _fit_kth_nn_distribution(nn_distances: dict, min_samples: int, resolution: int = 200, verbose=True):
    """Fits a kth nn distribution model to a dict of arrays of nn distances."""
    result = {}

    if verbose:
        print("    fitting kth nn distributions!")

    for a_cluster in nn_distances:

        # Sort nn distances and make an array of point numbers
        x_data = np.sort(nn_distances[a_cluster])
        y_data = np.arange(1, len(nn_distances[a_cluster]) + 1)

        # Interpolate this function into something more valid for fitting
        y_interpolator = interp1d(x_data, y_data, kind="linear")

        x_range = np.linspace(x_data.min(), x_data.max(), num=resolution)
        y_range = y_interpolator(x_range)

        # Normalise
        y_range /= np.trapz(y_range, x=x_range)

        # Fit a function!
        # Get starting guesses
        k = min_samples
        d_0 = min_samples / 2  # Might want a better estimate sometimes!
        a_0 = constrained_a(d_0, k, np.mean(nn_distances[a_cluster]))

        # Minimize - we suppress warnings because it's quite common for the curve_fit method to go into disallowed areas
        # sometimes
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                minimisation_result, covariance = curve_fit(
                    _CurveToFit(k), x_range, y_range, p0=np.asarray([a_0, d_0]), method="trf",
                    bounds=([0, 0], [np.inf, np.inf]),
                    verbose=False)

        # Catch if the fitting fails
        except RuntimeError as e:
            warnings.warn("fitting failed for a cluster!", RuntimeWarning)
            minimisation_result = (np.nan, np.nan)

        result[a_cluster] = (minimisation_result[0], minimisation_result[1], k)

    # Return values
    return result


def _likelihood_of_object(x_data, a, d, k, cdf_resolution=200):
    # Get decent estimates of the range of the function we should work with
    limits = (0.0, x_data.max() * 2)

    # Make a cdf!
    func = KthNNDistributionCDF(a, d, k, limits=limits, grid_size=cdf_resolution)

    # Evaluate all the points at the cdf!
    return np.sum(np.log(func(x_data)))


def _convert_two_sided_p_value_to_z_score(p_value):
    """Convert to a p value, then to a sigma significance - using a weird formula because it gives me better floating
    precision, and dividing by 2 since a z test is implicitly one-sided. I quite like this explanation of why:
    https://stats.idre.ucla.edu/other/mult-pkg/faq/general/faq-what-are-the-differences-between-one-tailed-and-two-tailed-tests/
    Basically, in our one tailed test we're only interested in the likelihood of the cluster being better than the
    field. We don't care about the field being better than the cluster: either way, that's a Z=0.0 not worth our
    consideration as an OC candidate.

    This differs slightly from a z score because it never gives you the sign of a result, only the "sigma significance"
    of the result. Handling the sign of your p-value should be done a-priori. For instance, you could first convert your
    p-value to a one sided test and *then* perform this.
    """
    # np.abs is to stop -0.0 from happening. Only 0.0 is allowed!
    return np.abs(-norm.ppf(p_value / 2))


def _convert_one_sided_p_value_to_z_score(p_value):
    """Converts a one-sided p value to a z score. Clipped so that p values greater than 0.5 will always return 0, since
    in that case there is effectively negative evidence that the alternate hypothesis is true.

    This is a nice explainer: https://www.youtube.com/watch?v=Z_pwYU5FVIs
    """
    return np.clip(-norm.ppf(p_value), 0.0, np.inf)


def _likelihood_ratio_test(nn_distances, cluster_fit_params, field_fit_params,
                           cdf_resolution=200):
    """Finds the maximum likelihood of the cluser & field models and does a simple comparison between the two
    to see which one is a better fit.

    See: http://rnowling.github.io/machine/learning/2017/10/07/likelihood-ratio-test.html

    """
    significances, log_likelihoods = {}, {}

    for a_cluster in nn_distances:

        # Grab likelihoods
        likelihood_cluster = _likelihood_of_object(nn_distances[a_cluster], *cluster_fit_params[a_cluster],
                                                   cdf_resolution=cdf_resolution)
        likelihood_field = _likelihood_of_object(nn_distances[a_cluster], *field_fit_params[a_cluster],
                                                 cdf_resolution=cdf_resolution)

        # Log and square the ratio!
        log_likelihoods[a_cluster] = 2 * (likelihood_cluster - likelihood_field)
        p_val = chi2.sf(log_likelihoods[a_cluster],
                        (len(nn_distances[a_cluster]) - 1) * (len(cluster_fit_params[a_cluster]) - 1))
        significances[a_cluster] = _convert_two_sided_p_value_to_z_score(p_val)

    return significances, log_likelihoods


def _ks_test(cluster_nn_distances: dict, field_nn_distances: dict, one_sample: bool = True,
             alternative: str = "greater", cdf_resolution: int = 200):
    """Performs a KS test (one-sided) of cluster stars and field stars to see if they're compatible or not.

    Args:
        cluster_nn_distances (np.ndarray): dict of arrays of cluster nearest neighbour distances of shape (n_stars,)
        field_nn_distances (np.ndarray): dict of arrays of len(3) of field fit parameters if one_sample==True, or dict
            of arrays of of field nearest neighbour distances otherwise.
        one_sample (bool): whether or not to do the test in one sample mode. Can produce slightly more reliable results
            (since the number of field stars will stop being a factor) but requires that the fitting procedure was
            correct.
            Default: True
        alternative (str): nature of the alternative hypothesis. May be one of ‘two-sided’, ‘less’ or ‘greater’. Greater
            will test that the cluster is *more tightly* distributed than the field, while less will test if it is
            *less tightly* distributed.
            Default: 'greater' (you shouldn't need anything else other than in exceptional circumstances)
        cdf_resolution (int): resolution to sample the cdf at if one_sample=True.
            Default: 200

    Returns:
        significance value of the clusters and the KS test statistics in a dict.

    """
    significances, ks_test_statistics = {}, {}

    for a_cluster in cluster_nn_distances:
        # Turn the field_nn_distances into a callable function instead of a CDF if one_sample==True.
        if one_sample:
            limits = (0.0, np.max(cluster_nn_distances[a_cluster]) * 2)
            field_distribution = KthNNDistributionCDF(
                *field_nn_distances[a_cluster], grid_size=cdf_resolution, limits=limits)
        else:
            field_distribution = field_nn_distances[a_cluster]

        # Do the KS test, get a p-value, go home
        ks_test_statistics[a_cluster], p_value = kstest(
            cluster_nn_distances[a_cluster], field_distribution, alternative=alternative)

        significances[a_cluster] = _convert_one_sided_p_value_to_z_score(p_value)

    return significances, ks_test_statistics


def _welch_t_test(cluster_nn_distances: dict, field_nn_distances: dict):
    """Compares the mean value of two nearest neighbour distributions to see if the cluster one is compatible with
    being more tightly distributed than the field one."""
    significances, t_test_statistics = {}, {}

    for a_cluster in cluster_nn_distances:
        # Calculate the means, stds, etc
        mean_c, std_c = np.mean(cluster_nn_distances[a_cluster]), np.std(cluster_nn_distances[a_cluster])
        mean_f, std_f = np.mean(field_nn_distances[a_cluster]), np.std(field_nn_distances[a_cluster])

        # We have to do a bit of work ourselves since scipy only has two-tailed t tests.
        # Immediately reject if mean_c > mean_f, i.e. the cluster is never gonna be more tightly distributed than f
        if mean_c > mean_f:
            significances[a_cluster], t_test_statistics[a_cluster] = 0.0, np.nan

        # Otherwise, we calculate the t test
        else:
            t_test_statistics[a_cluster], p_value = ttest_ind_from_stats(
                mean_c, std_c, len(cluster_nn_distances[a_cluster]),
                mean_f, std_f, len(field_nn_distances[a_cluster]),
                equal_var=False
            )
            significances[a_cluster] = _convert_two_sided_p_value_to_z_score(p_value)

    return significances, t_test_statistics


def _mann_whitney_rank_test(cluster_nn_distances: dict, field_nn_distances: dict):
    """Computes the Mann-Whitney rank test on the cluster and field distributions. Can be more robust than the t-test
    for non-Gaussian data.
    """
    significances, mw_test_statistics = {}, {}

    for a_cluster in cluster_nn_distances:
        # Calculate the test!
        mw_test_statistics[a_cluster], p_value = mannwhitneyu(
            cluster_nn_distances[a_cluster], field_nn_distances[a_cluster], alternative="less")
        significances[a_cluster] = _convert_one_sided_p_value_to_z_score(p_value)

    return significances, mw_test_statistics



"""A set of tests for use with the pytest module, covering ocelot.crossmatch"""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

location_mwsc = Path("./test_data/mwsc_ii_catalogue")
location_crossmatch_match_positions = Path("./test_data/crossmatch_match_position_only")
location_crossmatch_summary_positions = Path("./test_data/crossmatch_summary_position_only")
location_crossmatch_match_all = Path("./test_data/crossmatch_match_all_axes")
location_crossmatch_summary_all = Path("./test_data/crossmatch_summary_all_axes")


def test_catalogue():
    """Tests ocelot.crossmatch.Catalogue, the primary tool for cross-matching open clusters by comparing them against
    a known literature catalogue with arbitrary matchable parameters.

    """
    # Read in the catalogues
    list_of_10_ocs = ['Blanco_1', 'Stock_10', 'Kharchenko_1', 'NGC_2516', 'Trumpler_19',
                      'ASCC_79', 'Hogg_19', 'NGC_6475', 'Ruprecht_172', 'King_9']

    data_mwsc = pd.read_csv(location_mwsc)

    # ... and make a perturbed version.
    # All clusters have been lightly perturbed in position space and proper motion space. However, cluster 5 (ASCC 79)
    # has been offset in RA by 2 degrees, and cluster 9 (King 9) has been offset by 4kpc, making both of these clusters
    # into outliers.
    data_perturbed = data_mwsc[data_mwsc["Name"].isin(list_of_10_ocs)].sort_values("MWSC").reset_index().copy()

    np.random.seed(42)
    data_perturbed[["RAJ2000", "DEJ2000"]] += np.random.normal(loc=0.0, scale=0.001, size=(10, 2))
    data_perturbed[["pmRA", "pmDE"]] += np.random.normal(loc=0.0, scale=0.5, size=(10, 2))
    data_perturbed["d"] += np.random.normal(loc=0.0, scale=0.1, size=10) * data_perturbed["d"]
    data_perturbed.loc[5, "RAJ2000"] += 2
    data_perturbed.loc[9, "d"] -= 4000
    data_perturbed["ra_error"] = 0.001
    data_perturbed["dec_error"] = 0.001
    data_perturbed["distance_error"] = data_perturbed["d"] * 0.1

    key_override_dict = {"name": "Name",
                         "ra": "RAJ2000",
                         "dec": "DEJ2000",
                         "distance": "d",
                         "pmra": "pmRA",
                         "pmra_error": "e_pm",
                         "pmdec": "pmDE",
                         "pmdec_error": "e_pm",}

    # Create a catalogue with data_mwsc and try to crossmatch against it, but without any extra axes
    catalogue = ocelot.crossmatch.Catalogue(data_mwsc,
                                            "mwsc",
                                            key_name="Name",
                                            key_ra="RAJ2000",
                                            key_dec="DEJ2000",)

    catalogue.override_default_ocelot_parameter_names(key_override_dict)

    matches, summary = catalogue.crossmatch(data_perturbed, matches_to_record=2)

    # Test that both of the DataFrames for position only matching are correct (they were checked by hand originally)
    pd.testing.assert_frame_equal(matches, pd.read_csv(location_crossmatch_match_positions, index_col=0))
    pd.testing.assert_frame_equal(summary, pd.read_csv(location_crossmatch_summary_positions, index_col=0))

    # Create a catalogue with data_mwsc and try to crossmatch against it, now with many extra axes
    catalogue = ocelot.crossmatch.Catalogue(data_mwsc,
                                            "mwsc",
                                            key_name="Name",
                                            key_ra="RAJ2000",
                                            key_dec="DEJ2000",
                                            extra_axes=[["d", None],
                                                        ["pmRA", "e_pm"],
                                                        ["pmDE", "e_pm"]])

    catalogue.override_default_ocelot_parameter_names(key_override_dict)

    matches_all, summary_all = catalogue.crossmatch(data_perturbed,
                                                    extra_axes_keys=["distance", "pmra", "pmdec"],
                                                    matches_to_record=2)

    # Test that both of the DataFrames for all axis matching are correct (they were checked by hand originally)
    pd.testing.assert_frame_equal(matches_all, pd.read_csv(location_crossmatch_match_all, index_col=0))
    pd.testing.assert_frame_equal(summary_all, pd.read_csv(location_crossmatch_summary_all, index_col=0))

    return matches, summary, matches_all, summary_all, data_mwsc, data_perturbed


if __name__ == "__main__":
    pd.set_option('display.width', 160)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)
    match, summ, match1, summ1, mwsc, pert = test_catalogue()


